#!/usr/bin/env python

# Any copyright is dedicated to the Public Domain.
# http://creativecommons.org/publicdomain/zero/1.0/

import mozhttpd
import mozunit
import pytest
from six.moves.urllib.error import HTTPError
from six.moves.urllib.request import urlopen


def try_get(url, expected_contents):
    f = urlopen(url)
    assert f.getcode() == 200
    assert f.read() == expected_contents


def try_get_expect_404(url):
    with pytest.raises(HTTPError) as excinfo:
        urlopen(url)
    assert excinfo.value.code == 404


@pytest.fixture(name="httpd_basic")
def fixture_httpd_basic(tmpdir):
    d1 = tmpdir.mkdir("d1")
    d1.join("test1.txt").write("test 1 contents")

    d2 = tmpdir.mkdir("d2")
    d2.join("test2.txt").write("test 2 contents")

    httpd = mozhttpd.MozHttpd(
        port=0,
        docroot=str(d1),
        path_mappings={"/files": str(d2)},
    )
    httpd.start(block=False)

    yield httpd

    httpd.stop()
    d1.remove()
    d2.remove()


def test_basic(httpd_basic):
    """Test that requests to docroot and a path mapping work as expected."""
    try_get(httpd_basic.get_url("/test1.txt"), b"test 1 contents")
    try_get(httpd_basic.get_url("/files/test2.txt"), b"test 2 contents")
    try_get_expect_404(httpd_basic.get_url("/files/test2_nope.txt"))


@pytest.fixture(name="httpd_substring_mappings")
def fixture_httpd_substring_mappings(tmpdir):
    d1 = tmpdir.mkdir("d1")
    d1.join("test1.txt").write("test 1 contents")

    d2 = tmpdir.mkdir("d2")
    d2.join("test2.txt").write("test 2 contents")

    httpd = mozhttpd.MozHttpd(
        port=0,
        path_mappings={"/abcxyz": str(d1), "/abc": str(d2)},
    )
    httpd.start(block=False)
    yield httpd
    httpd.stop()
    d1.remove()
    d2.remove()


def test_substring_mappings(httpd_substring_mappings):
    httpd = httpd_substring_mappings
    try_get(httpd.get_url("/abcxyz/test1.txt"), b"test 1 contents")
    try_get(httpd.get_url("/abc/test2.txt"), b"test 2 contents")


@pytest.fixture(name="httpd_multipart_path_mapping")
def fixture_httpd_multipart_path_mapping(tmpdir):
    d1 = tmpdir.mkdir("d1")
    d1.join("test1.txt").write("test 1 contents")

    httpd = mozhttpd.MozHttpd(
        port=0,
        path_mappings={"/abc/def/ghi": str(d1)},
    )
    httpd.start(block=False)
    yield httpd
    httpd.stop()
    d1.remove()


def test_multipart_path_mapping(httpd_multipart_path_mapping):
    """Test that a path mapping with multiple directories works."""
    httpd = httpd_multipart_path_mapping
    try_get(httpd.get_url("/abc/def/ghi/test1.txt"), b"test 1 contents")
    try_get_expect_404(httpd.get_url("/abc/test1.txt"))
    try_get_expect_404(httpd.get_url("/abc/def/test1.txt"))


@pytest.fixture(name="httpd_no_docroot")
def fixture_httpd_no_docroot(tmpdir):
    d1 = tmpdir.mkdir("d1")
    httpd = mozhttpd.MozHttpd(
        port=0,
        path_mappings={"/foo": str(d1)},
    )
    httpd.start(block=False)
    yield httpd
    httpd.stop()
    d1.remove()


def test_no_docroot(httpd_no_docroot):
    """Test that path mappings with no docroot work."""
    try_get_expect_404(httpd_no_docroot.get_url())


if __name__ == "__main__":
    mozunit.main()



#!/usr/bin/env python3

# tooltool is a lookaside cache implemented in Python
# Copyright (C) 2011 John H. Ford <john@johnford.info>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation version 2
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

# A manifest file specifies files in that directory that are stored
# elsewhere. This file should only list files in the same directory
# in which the manifest file resides and it should be called
# 'manifest.tt'

import base64
import calendar
import hashlib
import hmac
import json
import logging
import math
import optparse
import os
import pprint
import re
import shutil
import stat
import sys
import tarfile
import tempfile
import threading
import time
import zipfile
from contextlib import closing, contextmanager
from functools import wraps
from io import BytesIO, open
from random import random
from subprocess import PIPE, Popen

__version__ = "1.4.0"

# Allowed request header characters:
# !#$%&'()*+,-./:;<=>?@[]^_`{|}~ and space, a-z, A-Z, 0-9, \, "
REQUEST_HEADER_ATTRIBUTE_CHARS = re.compile(
    r"^[ a-zA-Z0-9_\!#\$%&'\(\)\*\+,\-\./\:;<\=>\?@\[\]\^`\{\|\}~]*$"
)
DEFAULT_MANIFEST_NAME = "manifest.tt"
TOOLTOOL_PACKAGE_SUFFIX = ".TOOLTOOL-PACKAGE"
HAWK_VER = 1
PY3 = sys.version_info[0] == 3

if PY3:
    six_binary_type = bytes
    unicode = (
        str  # Silence `pyflakes` from reporting `undefined name 'unicode'` in Python 3.
    )
    import urllib.request as urllib2
    from http.client import HTTPConnection, HTTPSConnection
    from urllib.error import HTTPError, URLError
    from urllib.parse import urljoin, urlparse
    from urllib.request import Request
else:
    six_binary_type = str
    import urllib2
    from httplib import HTTPConnection, HTTPSConnection
    from urllib2 import HTTPError, Request, URLError
    from urlparse import urljoin, urlparse


log = logging.getLogger(__name__)


# Vendored code from `redo` module
def retrier(attempts=5, sleeptime=10, max_sleeptime=300, sleepscale=1.5, jitter=1):
    """
    This function originates from redo 2.0.3 https://github.com/mozilla-releng/redo
    A generator function that sleeps between retries, handles exponential
    backoff and jitter. The action you are retrying is meant to run after
    retrier yields.
    """
    jitter = jitter or 0  # py35 barfs on the next line if jitter is None
    if jitter > sleeptime:
        # To prevent negative sleep times
        raise Exception(
            "jitter ({}) must be less than sleep time ({})".format(jitter, sleeptime)
        )

    sleeptime_real = sleeptime
    for _ in range(attempts):
        log.debug("attempt %i/%i", _ + 1, attempts)

        yield sleeptime_real

        if jitter:
            sleeptime_real = sleeptime + random.uniform(-jitter, jitter)
            # our jitter should scale along with the sleeptime
            jitter = jitter * sleepscale
        else:
            sleeptime_real = sleeptime

        sleeptime *= sleepscale

        if sleeptime_real > max_sleeptime:
            sleeptime_real = max_sleeptime

        # Don't need to sleep the last time
        if _ < attempts - 1:
            log.debug(
                "sleeping for %.2fs (attempt %i/%i)", sleeptime_real, _ + 1, attempts
            )
            time.sleep(sleeptime_real)


def retry(
    action,
    attempts=5,
    sleeptime=60,
    max_sleeptime=5 * 60,
    sleepscale=1.5,
    jitter=1,
    retry_exceptions=(Exception,),
    cleanup=None,
    args=(),
    kwargs={},
    log_args=True,
):
    """
    This function originates from redo 2.0.3 https://github.com/mozilla-releng/redo
    Calls an action function until it succeeds, or we give up.
    """
    assert callable(action)
    assert not cleanup or callable(cleanup)

    action_name = getattr(action, "__name__", action)
    if log_args and (args or kwargs):
        log_attempt_args = (
            "retry: calling %s with args: %s," " kwargs: %s, attempt #%d",
            action_name,
            args,
            kwargs,
        )
    else:
        log_attempt_args = ("retry: calling %s, attempt #%d", action_name)

    if max_sleeptime < sleeptime:
        log.debug("max_sleeptime %d less than sleeptime %d", max_sleeptime, sleeptime)

    n = 1
    for _ in retrier(
        attempts=attempts,
        sleeptime=sleeptime,
        max_sleeptime=max_sleeptime,
        sleepscale=sleepscale,
        jitter=jitter,
    ):
        try:
            logfn = log.info if n != 1 else log.debug
            logfn_args = log_attempt_args + (n,)
            logfn(*logfn_args)
            return action(*args, **kwargs)
        except retry_exceptions:
            log.debug("retry: Caught exception: ", exc_info=True)
            if cleanup:
                cleanup()
            if n == attempts:
                log.info("retry: Giving up on %s", action_name)
                raise
            continue
        finally:
            n += 1


def retriable(*retry_args, **retry_kwargs):
    """
    This function originates from redo 2.0.3 https://github.com/mozilla-releng/redo
    A decorator factory for retry(). Wrap your function in @retriable(...) to
    give it retry powers!
    """

    def _retriable_factory(func):
        @wraps(func)
        def _retriable_wrapper(*args, **kwargs):
            return retry(func, args=args, kwargs=kwargs, *retry_args, **retry_kwargs)

        return _retriable_wrapper

    return _retriable_factory


# end of vendored code from redo module


def request_has_data(req):
    if PY3:
        return req.data is not None
    return req.has_data()


def get_hexdigest(val):
    return hashlib.sha512(val).hexdigest()


class FileRecordJSONEncoderException(Exception):
    pass


class InvalidManifest(Exception):
    pass


class ExceptionWithFilename(Exception):
    def __init__(self, filename):
        Exception.__init__(self)
        self.filename = filename


class BadFilenameException(ExceptionWithFilename):
    pass


class DigestMismatchException(ExceptionWithFilename):
    pass


class MissingFileException(ExceptionWithFilename):
    pass


class InvalidCredentials(Exception):
    pass


class BadHeaderValue(Exception):
    pass


def parse_url(url):
    url_parts = urlparse(url)
    url_dict = {
        "scheme": url_parts.scheme,
        "hostname": url_parts.hostname,
        "port": url_parts.port,
        "path": url_parts.path,
        "resource": url_parts.path,
        "query": url_parts.query,
    }
    if len(url_dict["query"]) > 0:
        url_dict["resource"] = "%s?%s" % (
            url_dict["resource"],  # pragma: no cover
            url_dict["query"],
        )

    if url_parts.port is None:
        if url_parts.scheme == "http":
            url_dict["port"] = 80
        elif url_parts.scheme == "https":  # pragma: no cover
            url_dict["port"] = 443
    return url_dict


def utc_now(offset_in_seconds=0.0):
    return int(math.floor(calendar.timegm(time.gmtime()) + float(offset_in_seconds)))


def random_string(length):
    return base64.urlsafe_b64encode(os.urandom(length))[:length]


def prepare_header_val(val):
    if isinstance(val, six_binary_type):
        val = val.decode("utf-8")

    if not REQUEST_HEADER_ATTRIBUTE_CHARS.match(val):
        raise BadHeaderValue(  # pragma: no cover
            "header value value={val} contained an illegal character".format(
                val=repr(val)
            )
        )

    return val


def parse_content_type(content_type):  # pragma: no cover
    if content_type:
        return content_type.split(";")[0].strip().lower()
    else:
        return ""


def calculate_payload_hash(algorithm, payload, content_type):  # pragma: no cover
    parts = [
        part if isinstance(part, six_binary_type) else part.encode("utf8")
        for part in [
            "hawk." + str(HAWK_VER) + ".payload\n",
            parse_content_type(content_type) + "\n",
            payload or "",
            "\n",
        ]
    ]

    p_hash = hashlib.new(algorithm)
    for p in parts:
        p_hash.update(p)

    log.debug(
        "calculating payload hash from:\n{parts}".format(parts=pprint.pformat(parts))
    )

    return base64.b64encode(p_hash.digest())


def validate_taskcluster_credentials(credentials):
    if not hasattr(credentials, "__getitem__"):
        raise InvalidCredentials(
            "credentials must be a dict-like object"
        )  # pragma: no cover
    try:
        credentials["clientId"]
        credentials["accessToken"]
    except KeyError:  # pragma: no cover
        etype, val, tb = sys.exc_info()
        raise InvalidCredentials("{etype}: {val}".format(etype=etype, val=val))


def normalize_header_attr(val):
    if isinstance(val, six_binary_type):
        return val.decode("utf-8")
    return val  # pragma: no cover


def normalize_string(
    mac_type,
    timestamp,
    nonce,
    method,
    name,
    host,
    port,
    content_hash,
):
    return "\n".join(
        [
            normalize_header_attr(header)
            # The blank lines are important. They follow what the Node Hawk lib does.
            for header in [
                "hawk." + str(HAWK_VER) + "." + mac_type,
                timestamp,
                nonce,
                method or "",
                name or "",
                host,
                port,
                content_hash or "",
                "",  # for ext which is empty in this case
                "",  # Add trailing new line.
            ]
        ]
    )


def calculate_mac(
    mac_type,
    access_token,
    algorithm,
    timestamp,
    nonce,
    method,
    name,
    host,
    port,
    content_hash,
):
    normalized = normalize_string(
        mac_type, timestamp, nonce, method, name, host, port, content_hash
    )
    log.debug(u"normalized resource for mac calc: {norm}".format(norm=normalized))
    digestmod = getattr(hashlib, algorithm)

    if not isinstance(normalized, six_binary_type):
        normalized = normalized.encode("utf8")

    if not isinstance(access_token, six_binary_type):
        access_token = access_token.encode("ascii")

    result = hmac.new(access_token, normalized, digestmod)
    return base64.b64encode(result.digest())


def make_taskcluster_header(credentials, req):
    validate_taskcluster_credentials(credentials)

    url = req.get_full_url()
    method = req.get_method()
    algorithm = "sha256"
    timestamp = str(utc_now())
    nonce = random_string(6)
    url_parts = parse_url(url)

    content_hash = None
    if request_has_data(req):
        if PY3:
            data = req.data
        else:
            data = req.get_data()
        content_hash = calculate_payload_hash(  # pragma: no cover
            algorithm,
            data,
            # maybe we should detect this from req.headers but we anyway expect json
            content_type="application/json",
        )

    mac = calculate_mac(
        "header",
        credentials["accessToken"],
        algorithm,
        timestamp,
        nonce,
        method,
        url_parts["resource"],
        url_parts["hostname"],
        str(url_parts["port"]),
        content_hash,
    )

    header = u'Hawk mac="{}"'.format(prepare_header_val(mac))

    if content_hash:  # pragma: no cover
        header = u'{}, hash="{}"'.format(header, prepare_header_val(content_hash))

    header = u'{header}, id="{id}", ts="{ts}", nonce="{nonce}"'.format(
        header=header,
        id=prepare_header_val(credentials["clientId"]),
        ts=prepare_header_val(timestamp),
        nonce=prepare_header_val(nonce),
    )

    log.debug("Hawk header for URL={} method={}: {}".format(url, method, header))

    return header


class FileRecord(object):
    def __init__(
        self,
        filename,
        size,
        digest,
        algorithm,
        unpack=False,
        version=None,
        visibility=None,
    ):
        object.__init__(self)
        if "/" in filename or "\\" in filename:
            log.error(
                "The filename provided contains path information and is, therefore, invalid."
            )
            raise BadFilenameException(filename=filename)
        self.filename = filename
        self.size = size
        self.digest = digest
        self.algorithm = algorithm
        self.unpack = unpack
        self.version = version
        self.visibility = visibility

    def __eq__(self, other):
        if self is other:
            return True
        if (
            self.filename == other.filename
            and self.size == other.size
            and self.digest == other.digest
            and self.algorithm == other.algorithm
            and self.version == other.version
            and self.visibility == other.visibility
        ):
            return True
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return (
            "%s.%s(filename='%s', size=%s, digest='%s', algorithm='%s', visibility=%r)"
            % (
                __name__,
                self.__class__.__name__,
                self.filename,
                self.size,
                self.digest,
                self.algorithm,
                self.visibility,
            )
        )

    def present(self):
        # Doesn't check validity
        return os.path.exists(self.filename)

    def validate_size(self):
        if self.present():
            return self.size == os.path.getsize(self.filename)
        else:
            log.debug("trying to validate size on a missing file, %s", self.filename)
            raise MissingFileException(filename=self.filename)

    def validate_digest(self):
        if self.present():
            with open(self.filename, "rb") as f:
                return self.digest == digest_file(f, self.algorithm)
        else:
            log.debug("trying to validate digest on a missing file, %s', self.filename")
            raise MissingFileException(filename=self.filename)

    def validate(self):
        if self.size is None or self.validate_size():
            if self.validate_digest():
                return True
        return False

    def describe(self):
        if self.present() and self.validate():
            return "'%s' is present and valid" % self.filename
        elif self.present():
            return "'%s' is present and invalid" % self.filename
        else:
            return "'%s' is absent" % self.filename


def create_file_record(filename, algorithm):
    fo = open(filename, "rb")
    stored_filename = os.path.split(filename)[1]
    fr = FileRecord(
        stored_filename,
        os.path.getsize(filename),
        digest_file(fo, algorithm),
        algorithm,
    )
    fo.close()
    return fr


class FileRecordJSONEncoder(json.JSONEncoder):
    def encode_file_record(self, obj):
        if not issubclass(type(obj), FileRecord):
            err = (
                "FileRecordJSONEncoder is only for FileRecord and lists of FileRecords, "
                "not %s" % obj.__class__.__name__
            )
            log.warn(err)
            raise FileRecordJSONEncoderException(err)
        else:
            rv = {
                "filename": obj.filename,
                "size": obj.size,
                "algorithm": obj.algorithm,
                "digest": obj.digest,
            }
            if obj.unpack:
                rv["unpack"] = True
            if obj.version:
                rv["version"] = obj.version
            if obj.visibility is not None:
                rv["visibility"] = obj.visibility
            return rv

    def default(self, f):
        if issubclass(type(f), list):
            record_list = []
            for i in f:
                record_list.append(self.encode_file_record(i))
            return record_list
        else:
            return self.encode_file_record(f)


class FileRecordJSONDecoder(json.JSONDecoder):

    """I help the json module materialize a FileRecord from
    a JSON file.  I understand FileRecords and lists of
    FileRecords.  I ignore things that I don't expect for now"""

    # TODO: make this more explicit in what it's looking for
    # and error out on unexpected things

    def process_file_records(self, obj):
        if isinstance(obj, list):
            record_list = []
            for i in obj:
                record = self.process_file_records(i)
                if issubclass(type(record), FileRecord):
                    record_list.append(record)
            return record_list
        required_fields = [
            "filename",
            "size",
            "algorithm",
            "digest",
        ]
        if isinstance(obj, dict):
            missing = False
            for req in required_fields:
                if req not in obj:
                    missing = True
                    break

            if not missing:
                unpack = obj.get("unpack", False)
                version = obj.get("version", None)
                visibility = obj.get("visibility", None)
                rv = FileRecord(
                    obj["filename"],
                    obj["size"],
                    obj["digest"],
                    obj["algorithm"],
                    unpack,
                    version,
                    visibility,
                )
                log.debug("materialized %s" % rv)
                return rv
        return obj

    def decode(self, s):
        decoded = json.JSONDecoder.decode(self, s)
        rv = self.process_file_records(decoded)
        return rv


class Manifest(object):

    valid_formats = ("json",)

    def __init__(self, file_records=None):
        self.file_records = file_records or []

    def __eq__(self, other):
        if self is other:
            return True
        if len(self.file_records) != len(other.file_records):
            log.debug("Manifests differ in number of files")
            return False
        # sort the file records by filename before comparing
        mine = sorted((fr.filename, fr) for fr in self.file_records)
        theirs = sorted((fr.filename, fr) for fr in other.file_records)
        return mine == theirs

    def __ne__(self, other):
        return not self.__eq__(other)

    def __deepcopy__(self, memo):
        # This is required for a deep copy
        return Manifest(self.file_records[:])

    def __copy__(self):
        return Manifest(self.file_records)

    def copy(self):
        return Manifest(self.file_records[:])

    def present(self):
        return all(i.present() for i in self.file_records)

    def validate_sizes(self):
        return all(i.validate_size() for i in self.file_records)

    def validate_digests(self):
        return all(i.validate_digest() for i in self.file_records)

    def validate(self):
        return all(i.validate() for i in self.file_records)

    def load(self, data_file, fmt="json"):
        assert fmt in self.valid_formats
        if fmt == "json":
            try:
                self.file_records.extend(
                    json.load(data_file, cls=FileRecordJSONDecoder)
                )
            except ValueError:
                raise InvalidManifest("trying to read invalid manifest file")

    def loads(self, data_string, fmt="json"):
        assert fmt in self.valid_formats
        if fmt == "json":
            try:
                self.file_records.extend(
                    json.loads(data_string, cls=FileRecordJSONDecoder)
                )
            except ValueError:
                raise InvalidManifest("trying to read invalid manifest file")

    def dump(self, output_file, fmt="json"):
        assert fmt in self.valid_formats
        if fmt == "json":
            return json.dump(
                self.file_records,
                output_file,
                indent=2,
                separators=(",", ": "),
                cls=FileRecordJSONEncoder,
            )

    def dumps(self, fmt="json"):
        assert fmt in self.valid_formats
        if fmt == "json":
            return json.dumps(
                self.file_records,
                indent=2,
                separators=(",", ": "),
                cls=FileRecordJSONEncoder,
            )


def digest_file(f, a):
    """I take a file like object 'f' and return a hex-string containing
    of the result of the algorithm 'a' applied to 'f'."""
    h = hashlib.new(a)
    chunk_size = 1024 * 10
    data = f.read(chunk_size)
    while data:
        h.update(data)
        data = f.read(chunk_size)
    name = repr(f.name) if hasattr(f, "name") else "a file"
    log.debug("hashed %s with %s to be %s", name, a, h.hexdigest())
    return h.hexdigest()


def execute(cmd):
    """Execute CMD, logging its stdout at the info level"""
    process = Popen(cmd, shell=True, stdout=PIPE)
    while True:
        line = process.stdout.readline()
        if not line:
            break
        log.info(line.replace("\n", " "))
    return process.wait() == 0


def open_manifest(manifest_file):
    """I know how to take a filename and load it into a Manifest object"""
    if os.path.exists(manifest_file):
        manifest = Manifest()
        with open(manifest_file, "r" if PY3 else "rb") as f:
            manifest.load(f)
            log.debug("loaded manifest from file '%s'" % manifest_file)
        return manifest
    else:
        log.debug("tried to load absent file '%s' as manifest" % manifest_file)
        raise InvalidManifest("manifest file '%s' does not exist" % manifest_file)


def list_manifest(manifest_file):
    """I know how print all the files in a location"""
    try:
        manifest = open_manifest(manifest_file)
    except InvalidManifest as e:
        log.error(
            "failed to load manifest file at '%s': %s"
            % (
                manifest_file,
                str(e),
            )
        )
        return False
    for f in manifest.file_records:
        print(
            "{}\t{}\t{}".format(
                "P" if f.present() else "-",
                "V" if f.present() and f.validate() else "-",
                f.filename,
            )
        )
    return True


def validate_manifest(manifest_file):
    """I validate that all files in a manifest are present and valid but
    don't fetch or delete them if they aren't"""
    try:
        manifest = open_manifest(manifest_file)
    except InvalidManifest as e:
        log.error(
            "failed to load manifest file at '%s': %s"
            % (
                manifest_file,
                str(e),
            )
        )
        return False
    invalid_files = []
    absent_files = []
    for f in manifest.file_records:
        if not f.present():
            absent_files.append(f)
        else:
            if not f.validate():
                invalid_files.append(f)
    if len(invalid_files + absent_files) == 0:
        return True
    else:
        return False


def add_files(manifest_file, algorithm, filenames, version, visibility, unpack):
    # returns True if all files successfully added, False if not
    # and doesn't catch library Exceptions.  If any files are already
    # tracked in the manifest, return will be False because they weren't
    # added
    all_files_added = True
    # Create a old_manifest object to add to
    if os.path.exists(manifest_file):
        old_manifest = open_manifest(manifest_file)
    else:
        old_manifest = Manifest()
        log.debug("creating a new manifest file")
    new_manifest = Manifest()  # use a different manifest for the output
    for filename in filenames:
        log.debug("adding %s" % filename)
        path, name = os.path.split(filename)
        new_fr = create_file_record(filename, algorithm)
        new_fr.version = version
        new_fr.visibility = visibility
        new_fr.unpack = unpack
        log.debug("appending a new file record to manifest file")
        add = True
        for fr in old_manifest.file_records:
            log.debug(
                "manifest file has '%s'"
                % "', ".join([x.filename for x in old_manifest.file_records])
            )
            if new_fr == fr:
                log.info("file already in old_manifest")
                add = False
            elif filename == fr.filename:
                log.error(
                    "manifest already contains a different file named %s" % filename
                )
                add = False
        if add:
            new_manifest.file_records.append(new_fr)
            log.debug("added '%s' to manifest" % filename)
        else:
            all_files_added = False
    # copy any files in the old manifest that aren't in the new one
    new_filenames = set(fr.filename for fr in new_manifest.file_records)
    for old_fr in old_manifest.file_records:
        if old_fr.filename not in new_filenames:
            new_manifest.file_records.append(old_fr)
    if PY3:
        with open(manifest_file, mode="w") as output:
            new_manifest.dump(output, fmt="json")
    else:
        with open(manifest_file, mode="wb") as output:
            new_manifest.dump(output, fmt="json")
    return all_files_added


def touch(f):
    """Used to modify mtime in cached files;
    mtime is used by the purge command"""
    try:
        os.utime(f, None)
    except OSError:
        log.warn("impossible to update utime of file %s" % f)


@contextmanager
@retriable(sleeptime=2)
def request(url, auth_file=None):
    req = Request(url)
    _authorize(req, auth_file)
    with closing(urllib2.urlopen(req)) as f:
        log.debug("opened %s for reading" % url)
        yield f


def fetch_file(base_urls, file_record, grabchunk=1024 * 4, auth_file=None, region=None):
    # A file which is requested to be fetched that exists locally will be
    # overwritten by this function
    fd, temp_path = tempfile.mkstemp(dir=os.getcwd())
    os.close(fd)
    fetched_path = None
    for base_url in base_urls:
        # Generate the URL for the file on the server side
        url = urljoin(base_url, "%s/%s" % (file_record.algorithm, file_record.digest))
        if region is not None:
            url += "?region=" + region

        log.info("Attempting to fetch from '%s'..." % base_url)

        # Well, the file doesn't exist locally.  Let's fetch it.
        try:
            with request(url, auth_file) as f, open(temp_path, mode="wb") as out:
                k = True
                size = 0
                while k:
                    # TODO: print statistics as file transfers happen both for info and to stop
                    # buildbot timeouts
                    indata = f.read(grabchunk)
                    out.write(indata)
                    size += len(indata)
                    if len(indata) == 0:
                        k = False
                log.info(
                    "File %s fetched from %s as %s"
                    % (file_record.filename, base_url, temp_path)
                )
                fetched_path = temp_path
                break
        except (URLError, HTTPError, ValueError):
            log.info(
                "...failed to fetch '%s' from %s" % (file_record.filename, base_url),
                exc_info=True,
            )
        except IOError:  # pragma: no cover
            log.info(
                "failed to write to temporary file for '%s'" % file_record.filename,
                exc_info=True,
            )

    # cleanup temp file in case of issues
    if fetched_path:
        return os.path.split(fetched_path)[1]
    else:
        try:
            os.remove(temp_path)
        except OSError:  # pragma: no cover
            pass
        return None


def clean_path(dirname):
    """Remove a subtree if is exists. Helper for unpack_file()."""
    if os.path.exists(dirname):
        log.info("rm tree: %s" % dirname)
        shutil.rmtree(dirname)


CHECKSUM_SUFFIX = ".checksum"


def validate_tar_member(member, path):
    def _is_within_directory(directory, target):
        real_directory = os.path.realpath(directory)
        real_target = os.path.realpath(target)
        prefix = os.path.commonprefix([real_directory, real_target])
        return prefix == real_directory

    member_path = os.path.join(path, member.name)
    if not _is_within_directory(path, member_path):
        raise Exception("Attempted path traversal in tar file: " + member.name)
    if member.issym():
        link_path = os.path.join(os.path.dirname(member_path), member.linkname)
        if not _is_within_directory(path, link_path):
            raise Exception("Attempted link path traversal in tar file: " + member.name)
    if member.mode & (stat.S_ISUID | stat.S_ISGID):
        raise Exception("Attempted setuid or setgid in tar file: " + member.name)


def safe_extract(tar, path=".", *, numeric_owner=False):
    def _files(tar, path):
        for member in tar:
            validate_tar_member(member, path)
            yield member

    tar.extractall(path, members=_files(tar, path), numeric_owner=numeric_owner)


def unpack_file(filename):
    """Untar `filename`, assuming it is uncompressed or compressed with bzip2,
    xz, gzip, zst, or unzip a zip file. The file is assumed to contain a single
    directory with a name matching the base of the given filename.
    Xz support is handled by shelling out to 'tar'."""
    if os.path.isfile(filename) and tarfile.is_tarfile(filename):
        tar_file, zip_ext = os.path.splitext(filename)
        base_file, tar_ext = os.path.splitext(tar_file)
        clean_path(base_file)
        log.info('untarring "%s"' % filename)
        with tarfile.open(filename) as tar:
            safe_extract(tar)
    elif os.path.isfile(filename) and filename.endswith(".tar.xz"):
        base_file = filename.replace(".tar.xz", "")
        clean_path(base_file)
        log.info('untarring "%s"' % filename)
        # Not using tar -Jxf because it fails on Windows for some reason.
        process = Popen(["xz", "-d", "-c", filename], stdout=PIPE)
        stdout, stderr = process.communicate()
        if process.returncode != 0:
            return False
        fileobj = BytesIO()
        fileobj.write(stdout)
        fileobj.seek(0)
        with tarfile.open(fileobj=fileobj, mode="r|") as tar:
            safe_extract(tar)
    elif os.path.isfile(filename) and filename.endswith(".tar.zst"):
        import zstandard

        base_file = filename.replace(".tar.zst", "")
        clean_path(base_file)
        log.info('untarring "%s"' % filename)
        dctx = zstandard.ZstdDecompressor()
        with dctx.stream_reader(open(filename, "rb")) as fileobj:
            with tarfile.open(fileobj=fileobj, mode="r|") as tar:
                safe_extract(tar)
    elif os.path.isfile(filename) and zipfile.is_zipfile(filename):
        base_file = filename.replace(".zip", "")
        clean_path(base_file)
        log.info('unzipping "%s"' % filename)
        z = zipfile.ZipFile(filename)
        z.extractall()
        z.close()
    else:
        log.error("Unknown archive extension for filename '%s'" % filename)
        return False
    return True


def fetch_files(
    manifest_file,
    base_urls,
    filenames=[],
    cache_folder=None,
    auth_file=None,
    region=None,
):
    # Lets load the manifest file
    try:
        manifest = open_manifest(manifest_file)
    except InvalidManifest as e:
        log.error(
            "failed to load manifest file at '%s': %s"
            % (
                manifest_file,
                str(e),
            )
        )
        return False

    # we want to track files already in current working directory AND valid
    # we will not need to fetch these
    present_files = []

    # We want to track files that fail to be fetched as well as
    # files that are fetched
    failed_files = []
    fetched_files = []

    # Files that we want to unpack.
    unpack_files = []

    # Lets go through the manifest and fetch the files that we want
    for f in manifest.file_records:
        # case 1: files are already present
        if f.present():
            if f.validate():
                present_files.append(f.filename)
                if f.unpack:
                    unpack_files.append(f.filename)
            else:
                # we have an invalid file here, better to cleanup!
                # this invalid file needs to be replaced with a good one
                # from the local cash or fetched from a tooltool server
                log.info(
                    "File %s is present locally but it is invalid, so I will remove it "
                    "and try to fetch it" % f.filename
                )
                os.remove(os.path.join(os.getcwd(), f.filename))

        # check if file is already in cache
        if cache_folder and f.filename not in present_files:
            try:
                shutil.copy(
                    os.path.join(cache_folder, f.digest),
                    os.path.join(os.getcwd(), f.filename),
                )
                log.info(
                    "File %s retrieved from local cache %s" % (f.filename, cache_folder)
                )
                touch(os.path.join(cache_folder, f.digest))

                filerecord_for_validation = FileRecord(
                    f.filename, f.size, f.digest, f.algorithm
                )
                if filerecord_for_validation.validate():
                    present_files.append(f.filename)
                    if f.unpack:
                        unpack_files.append(f.filename)
                else:
                    # the file copied from the cache is invalid, better to
                    # clean up the cache version itself as well
                    log.warn(
                        "File %s retrieved from cache is invalid! I am deleting it from the "
                        "cache as well" % f.filename
                    )
                    os.remove(os.path.join(os.getcwd(), f.filename))
                    os.remove(os.path.join(cache_folder, f.digest))
            except IOError:
                log.info(
                    "File %s not present in local cache folder %s"
                    % (f.filename, cache_folder)
                )

        # now I will try to fetch all files which are not already present and
        # valid, appending a suffix to avoid race conditions
        temp_file_name = None
        # 'filenames' is the list of filenames to be managed, if this variable
        # is a non empty list it can be used to filter if filename is in
        # present_files, it means that I have it already because it was already
        # either in the working dir or in the cache
        if (
            f.filename in filenames or len(filenames) == 0
        ) and f.filename not in present_files:
            log.debug("fetching %s" % f.filename)
            temp_file_name = fetch_file(
                base_urls, f, auth_file=auth_file, region=region
            )
            if temp_file_name:
                fetched_files.append((f, temp_file_name))
            else:
                failed_files.append(f.filename)
        else:
            log.debug("skipping %s" % f.filename)

    # lets ensure that fetched files match what the manifest specified
    for localfile, temp_file_name in fetched_files:
        # since I downloaded to a temp file, I need to perform all validations on the temp file
        # this is why filerecord_for_validation is created

        filerecord_for_validation = FileRecord(
            temp_file_name, localfile.size, localfile.digest, localfile.algorithm
        )

        if filerecord_for_validation.validate():
            # great!
            # I can rename the temp file
            log.info(
                "File integrity verified, renaming %s to %s"
                % (temp_file_name, localfile.filename)
            )
            os.rename(
                os.path.join(os.getcwd(), temp_file_name),
                os.path.join(os.getcwd(), localfile.filename),
            )

            if localfile.unpack:
                unpack_files.append(localfile.filename)

            # if I am using a cache and a new file has just been retrieved from a
            # remote location, I need to update the cache as well
            if cache_folder:
                log.info("Updating local cache %s..." % cache_folder)
                try:
                    if not os.path.exists(cache_folder):
                        log.info("Creating cache in %s..." % cache_folder)
                        os.makedirs(cache_folder, 0o0700)
                    shutil.copy(
                        os.path.join(os.getcwd(), localfile.filename),
                        os.path.join(cache_folder, localfile.digest),
                    )
                    log.info(
                        "Local cache %s updated with %s"
                        % (cache_folder, localfile.filename)
                    )
                    touch(os.path.join(cache_folder, localfile.digest))
                except (OSError, IOError):
                    log.warning(
                        "Impossible to add file %s to cache folder %s"
                        % (localfile.filename, cache_folder),
                        exc_info=True,
                    )
        else:
            failed_files.append(localfile.filename)
            log.error("'%s'" % filerecord_for_validation.describe())
            os.remove(temp_file_name)

    # Unpack files that need to be unpacked.
    for filename in unpack_files:
        if not unpack_file(filename):
            failed_files.append(filename)

    # If we failed to fetch or validate a file, we need to fail
    if len(failed_files) > 0:
        log.error("The following files failed: '%s'" % "', ".join(failed_files))
        return False
    return True


def freespace(p):
    "Returns the number of bytes free under directory `p`"
    if sys.platform == "win32":  # pragma: no cover
        # os.statvfs doesn't work on Windows
        import win32file

        secsPerClus, bytesPerSec, nFreeClus, totClus = win32file.GetDiskFreeSpace(p)
        return secsPerClus * bytesPerSec * nFreeClus
    else:
        r = os.statvfs(p)
        return r.f_frsize * r.f_bavail


def purge(folder, gigs):
    """If gigs is non 0, it deletes files in `folder` until `gigs` GB are free,
    starting from older files.  If gigs is 0, a full purge will be performed.
    No recursive deletion of files in subfolder is performed."""

    full_purge = bool(gigs == 0)
    gigs *= 1024 * 1024 * 1024

    if not full_purge and freespace(folder) >= gigs:
        log.info("No need to cleanup")
        return

    files = []
    for f in os.listdir(folder):
        p = os.path.join(folder, f)
        # it deletes files in folder without going into subfolders,
        # assuming the cache has a flat structure
        if not os.path.isfile(p):
            continue
        mtime = os.path.getmtime(p)
        files.append((mtime, p))

    # iterate files sorted by mtime
    for _, f in sorted(files):
        log.info("removing %s to free up space" % f)
        try:
            os.remove(f)
        except OSError:
            log.info("Impossible to remove %s" % f, exc_info=True)
        if not full_purge and freespace(folder) >= gigs:
            break


def _log_api_error(e):
    if hasattr(e, "hdrs") and e.hdrs["content-type"] == "application/json":
        json_resp = json.load(e.fp)
        log.error(
            "%s: %s" % (json_resp["error"]["name"], json_resp["error"]["description"])
        )
    else:
        log.exception("Error making RelengAPI request:")


def _authorize(req, auth_file):
    is_taskcluster_auth = False

    if not auth_file:
        try:
            taskcluster_env_keys = {
                "clientId": "TASKCLUSTER_CLIENT_ID",
                "accessToken": "TASKCLUSTER_ACCESS_TOKEN",
            }
            auth_content = {k: os.environ[v] for k, v in taskcluster_env_keys.items()}
            is_taskcluster_auth = True
        except KeyError:
            return
    else:
        with open(auth_file) as f:
            auth_content = f.read().strip()
            try:
                auth_content = json.loads(auth_content)
                is_taskcluster_auth = True
            except Exception:
                pass

    if is_taskcluster_auth:
        taskcluster_header = make_taskcluster_header(auth_content, req)
        log.debug("Using taskcluster credentials in %s" % auth_file)
        req.add_unredirected_header("Authorization", taskcluster_header)
    else:
        log.debug("Using Bearer token in %s" % auth_file)
        req.add_unredirected_header("Authorization", "Bearer %s" % auth_content)


def _send_batch(base_url, auth_file, batch, region):
    url = urljoin(base_url, "upload")
    if region is not None:
        url += "?region=" + region
    data = json.dumps(batch)
    if PY3:
        data = data.encode("utf-8")
    req = Request(url, data, {"Content-Type": "application/json"})
    _authorize(req, auth_file)
    try:
        resp = urllib2.urlopen(req)
    except (URLError, HTTPError) as e:
        _log_api_error(e)
        return None
    return json.load(resp)["result"]


def _s3_upload(filename, file):
    # urllib2 does not support streaming, so we fall back to good old httplib
    url = urlparse(file["put_url"])
    cls = HTTPSConnection if url.scheme == "https" else HTTPConnection
    host, port = url.netloc.split(":") if ":" in url.netloc else (url.netloc, 443)
    port = int(port)
    conn = cls(host, port)
    try:
        req_path = "%s?%s" % (url.path, url.query) if url.query else url.path
        with open(filename, "rb") as f:
            content = f.read()
            content_length = len(content)
            f.seek(0)
            conn.request(
                "PUT",
                req_path,
                f,
                {
                    "Content-Type": "application/octet-stream",
                    "Content-Length": str(content_length),
                },
            )
            resp = conn.getresponse()
            resp_body = resp.read()
            conn.close()
        if resp.status != 200:
            raise RuntimeError(
                "Non-200 return from AWS: %s %s\n%s"
                % (resp.status, resp.reason, resp_body)
            )
    except Exception:
        file["upload_exception"] = sys.exc_info()
        file["upload_ok"] = False
    else:
        file["upload_ok"] = True


def _notify_upload_complete(base_url, auth_file, file):
    req = Request(urljoin(base_url, "upload/complete/%(algorithm)s/%(digest)s" % file))
    _authorize(req, auth_file)
    try:
        urllib2.urlopen(req)
    except HTTPError as e:
        if e.code != 409:
            _log_api_error(e)
            return
        # 409 indicates that the upload URL hasn't expired yet and we
        # should retry after a delay
        to_wait = int(e.headers.get("X-Retry-After", 60))
        log.warning("Waiting %d seconds for upload URLs to expire" % to_wait)
        time.sleep(to_wait)
        _notify_upload_complete(base_url, auth_file, file)
    except Exception:
        log.exception("While notifying server of upload completion:")


def upload(manifest, message, base_urls, auth_file, region):
    try:
        manifest = open_manifest(manifest)
    except InvalidManifest:
        log.exception("failed to load manifest file at '%s'")
        return False

    # verify the manifest, since we'll need the files present to upload
    if not manifest.validate():
        log.error("manifest is invalid")
        return False

    if any(fr.visibility is None for fr in manifest.file_records):
        log.error("All files in a manifest for upload must have a visibility set")

    # convert the manifest to an upload batch
    batch = {
        "message": message,
        "files": {},
    }
    for fr in manifest.file_records:
        batch["files"][fr.filename] = {
            "size": fr.size,
            "digest": fr.digest,
            "algorithm": fr.algorithm,
            "visibility": fr.visibility,
        }

    # make the upload request
    resp = _send_batch(base_urls[0], auth_file, batch, region)
    if not resp:
        return None
    files = resp["files"]

    # Upload the files, each in a thread.  This allows us to start all of the
    # uploads before any of the URLs expire.
    threads = {}
    for filename, file in files.items():
        if "put_url" in file:
            log.info("%s: starting upload" % (filename,))
            thd = threading.Thread(target=_s3_upload, args=(filename, file))
            thd.daemon = 1
            thd.start()
            threads[filename] = thd
        else:
            log.info("%s: already exists on server" % (filename,))

    # re-join all of those threads as they exit
    success = True
    while threads:
        for filename, thread in list(threads.items()):
            if not thread.is_alive():
                # _s3_upload has annotated file with result information
                file = files[filename]
                thread.join()
                if file["upload_ok"]:
                    log.info("%s: uploaded" % filename)
                else:
                    log.error(
                        "%s: failed" % filename, exc_info=file["upload_exception"]
                    )
                    success = False
                del threads[filename]

    # notify the server that the uploads are completed.  If the notification
    # fails, we don't consider that an error (the server will notice
    # eventually)
    for filename, file in files.items():
        if "put_url" in file and file["upload_ok"]:
            log.info("notifying server of upload completion for %s" % (filename,))
            _notify_upload_complete(base_urls[0], auth_file, file)

    return success


def send_operation_on_file(data, base_urls, digest, auth_file):
    url = base_urls[0]
    url = urljoin(url, "file/sha512/" + digest)

    data = json.dumps(data)

    req = Request(url, data, {"Content-Type": "application/json"})
    req.get_method = lambda: "PATCH"

    _authorize(req, auth_file)

    try:
        urllib2.urlopen(req)
    except (URLError, HTTPError) as e:
        _log_api_error(e)
        return False
    return True


def change_visibility(base_urls, digest, visibility, auth_file):
    data = [
        {
            "op": "set_visibility",
            "visibility": visibility,
        }
    ]
    return send_operation_on_file(data, base_urls, digest, auth_file)


def delete_instances(base_urls, digest, auth_file):
    data = [
        {
            "op": "delete_instances",
        }
    ]
    return send_operation_on_file(data, base_urls, digest, auth_file)


def process_command(options, args):
    """I know how to take a list of program arguments and
    start doing the right thing with them"""
    cmd = args[0]
    cmd_args = args[1:]
    log.debug("processing '%s' command with args '%s'" % (cmd, '", "'.join(cmd_args)))
    log.debug("using options: %s" % options)

    if cmd == "list":
        return list_manifest(options["manifest"])
    if cmd == "validate":
        return validate_manifest(options["manifest"])
    elif cmd == "add":
        return add_files(
            options["manifest"],
            options["algorithm"],
            cmd_args,
            options["version"],
            options["visibility"],
            options["unpack"],
        )
    elif cmd == "purge":
        if options["cache_folder"]:
            purge(folder=options["cache_folder"], gigs=options["size"])
        else:
            log.critical("please specify the cache folder to be purged")
            return False
    elif cmd == "fetch":
        return fetch_files(
            options["manifest"],
            options["base_url"],
            cmd_args,
            cache_folder=options["cache_folder"],
            auth_file=options.get("auth_file"),
            region=options.get("region"),
        )
    elif cmd == "upload":
        if not options.get("message"):
            log.critical("upload command requires a message")
            return False
        return upload(
            options.get("manifest"),
            options.get("message"),
            options.get("base_url"),
            options.get("auth_file"),
            options.get("region"),
        )
    elif cmd == "change-visibility":
        if not options.get("digest"):
            log.critical("change-visibility command requires a digest option")
            return False
        if not options.get("visibility"):
            log.critical("change-visibility command requires a visibility option")
            return False
        return change_visibility(
            options.get("base_url"),
            options.get("digest"),
            options.get("visibility"),
            options.get("auth_file"),
        )
    elif cmd == "delete":
        if not options.get("digest"):
            log.critical("change-visibility command requires a digest option")
            return False
        return delete_instances(
            options.get("base_url"),
            options.get("digest"),
            options.get("auth_file"),
        )
    else:
        log.critical('command "%s" is not implemented' % cmd)
        return False


def main(argv, _skip_logging=False):
    # Set up option parsing
    parser = optparse.OptionParser()
    parser.add_option(
        "-q",
        "--quiet",
        default=logging.INFO,
        dest="loglevel",
        action="store_const",
        const=logging.ERROR,
    )
    parser.add_option(
        "-v", "--verbose", dest="loglevel", action="store_const", const=logging.DEBUG
    )
    parser.add_option(
        "-m",
        "--manifest",
        default=DEFAULT_MANIFEST_NAME,
        dest="manifest",
        action="store",
        help="specify the manifest file to be operated on",
    )
    parser.add_option(
        "-d",
        "--algorithm",
        default="sha512",
        dest="algorithm",
        action="store",
        help="hashing algorithm to use (only sha512 is allowed)",
    )
    parser.add_option(
        "--digest",
        default=None,
        dest="digest",
        action="store",
        help="digest hash to change visibility for",
    )
    parser.add_option(
        "--visibility",
        default=None,
        dest="visibility",
        choices=["internal", "public"],
        help='Visibility level of this file; "internal" is for '
        "files that cannot be distributed out of the company "
        'but not for secrets; "public" files are available to '
        "anyone without restriction",
    )
    parser.add_option(
        "--unpack",
        default=False,
        dest="unpack",
        action="store_true",
        help="Request unpacking this file after fetch."
        " This is helpful with tarballs.",
    )
    parser.add_option(
        "--version",
        default=None,
        dest="version",
        action="store",
        help="Version string for this file. This annotates the "
        "manifest entry with a version string to help "
        "identify the contents.",
    )
    parser.add_option(
        "-o",
        "--overwrite",
        default=False,
        dest="overwrite",
        action="store_true",
        help="UNUSED; present for backward compatibility",
    )
    parser.add_option(
        "--url",
        dest="base_url",
        action="append",
        help="RelengAPI URL ending with /tooltool/; default "
        "is appropriate for Mozilla",
    )
    parser.add_option(
        "-c", "--cache-folder", dest="cache_folder", help="Local cache folder"
    )
    parser.add_option(
        "-s",
        "--size",
        help="free space required (in GB)",
        dest="size",
        type="float",
        default=0.0,
    )
    parser.add_option(
        "-r",
        "--region",
        help="Preferred AWS region for upload or fetch; " "example: --region=us-west-2",
    )
    parser.add_option(
        "--message",
        help='The "commit message" for an upload; format with a bug number '
        "and brief comment",
        dest="message",
    )
    parser.add_option(
        "--authentication-file",
        help="Use the RelengAPI token found in the given file to "
        "authenticate to the RelengAPI server.",
        dest="auth_file",
    )

    (options_obj, args) = parser.parse_args(argv[1:])

    if not options_obj.base_url:
        tooltool_host = os.environ.get("TOOLTOOL_HOST", "tooltool.mozilla-releng.net")
        taskcluster_proxy_url = os.environ.get("TASKCLUSTER_PROXY_URL")
        if taskcluster_proxy_url:
            tooltool_url = "{}/{}".format(taskcluster_proxy_url, tooltool_host)
        else:
            tooltool_url = "https://{}".format(tooltool_host)

        options_obj.base_url = [tooltool_url]

    # ensure all URLs have a trailing slash
    def add_slash(url):
        return url if url.endswith("/") else (url + "/")

    options_obj.base_url = [add_slash(u) for u in options_obj.base_url]

    # expand ~ in --authentication-file
    if options_obj.auth_file:
        options_obj.auth_file = os.path.expanduser(options_obj.auth_file)

    # Dictionaries are easier to work with
    options = vars(options_obj)

    log.setLevel(options["loglevel"])

    # Set up logging, for now just to the console
    if not _skip_logging:  # pragma: no cover
        ch = logging.StreamHandler()
        cf = logging.Formatter("%(levelname)s - %(message)s")
        ch.setFormatter(cf)
        log.addHandler(ch)

    if options["algorithm"] != "sha512":
        parser.error("only --algorithm sha512 is supported")

    if len(args) < 1:
        parser.error("You must specify a command")

    return 0 if process_command(options, args) else 1


if __name__ == "__main__":  # pragma: no cover
    sys.exit(main(sys.argv))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from six.moves.urllib.parse import quote

from marionette_driver.by import By
from marionette_harness import MarionetteTestCase, WindowManagerMixin


def inline(doc):
    return "data:text/html;charset=utf-8,{}".format(quote(doc))


class TestCloseWindow(WindowManagerMixin, MarionetteTestCase):
    def tearDown(self):
        self.close_all_windows()
        self.close_all_tabs()

        super(TestCloseWindow, self).tearDown()

    def test_close_chrome_window_for_browser_window(self):
        with self.marionette.using_context("chrome"):
            new_window = self.open_window()
        self.marionette.switch_to_window(new_window)

        self.assertIn(new_window, self.marionette.chrome_window_handles)
        chrome_window_handles = self.marionette.close_chrome_window()
        self.assertNotIn(new_window, chrome_window_handles)
        self.assertListEqual(self.start_windows, chrome_window_handles)
        self.assertNotIn(new_window, self.marionette.window_handles)

    def test_close_chrome_window_for_non_browser_window(self):
        new_window = self.open_chrome_window(
            "chrome://remote/content/marionette/test.xhtml"
        )
        self.marionette.switch_to_window(new_window)

        self.assertIn(new_window, self.marionette.chrome_window_handles)
        self.assertNotIn(new_window, self.marionette.window_handles)
        chrome_window_handles = self.marionette.close_chrome_window()
        self.assertNotIn(new_window, chrome_window_handles)
        self.assertListEqual(self.start_windows, chrome_window_handles)
        self.assertNotIn(new_window, self.marionette.window_handles)

    def test_close_chrome_window_for_last_open_window(self):
        self.close_all_windows()

        self.assertListEqual([], self.marionette.close_chrome_window())
        self.assertListEqual([self.start_tab], self.marionette.window_handles)
        self.assertListEqual([self.start_window], self.marionette.chrome_window_handles)
        self.assertIsNotNone(self.marionette.session)

    def test_close_window_for_browser_tab(self):
        new_tab = self.open_tab()
        self.marionette.switch_to_window(new_tab)

        window_handles = self.marionette.close()
        self.assertNotIn(new_tab, window_handles)
        self.assertListEqual(self.start_tabs, window_handles)

    def test_close_window_for_browser_window_with_single_tab(self):
        new_tab = self.open_window()
        self.marionette.switch_to_window(new_tab)

        self.assertEqual(len(self.marionette.window_handles), len(self.start_tabs) + 1)
        window_handles = self.marionette.close()
        self.assertNotIn(new_tab, window_handles)
        self.assertListEqual(self.start_tabs, window_handles)
        self.assertListEqual(self.start_windows, self.marionette.chrome_window_handles)

    def test_close_window_for_last_open_tab(self):
        self.close_all_tabs()

        self.assertListEqual([], self.marionette.close())
        self.assertListEqual([self.start_tab], self.marionette.window_handles)
        self.assertListEqual([self.start_window], self.marionette.chrome_window_handles)
        self.assertIsNotNone(self.marionette.session)

    def test_close_browserless_tab(self):
        self.close_all_tabs()

        test_page = self.marionette.absolute_url("windowHandles.html")
        new_tab = self.open_tab()
        self.marionette.switch_to_window(new_tab)
        self.marionette.navigate(test_page)
        self.marionette.switch_to_window(self.start_tab)

        with self.marionette.using_context("chrome"):
            self.marionette.execute_async_script(
                """
              const { BrowserWindowTracker } = ChromeUtils.importESModule(
                "resource:///modules/BrowserWindowTracker.sys.mjs"
              );

              let win = BrowserWindowTracker.getTopWindow();
              win.addEventListener("TabBrowserDiscarded", ev => {
                arguments[0](true);
              }, { once: true});
              win.gBrowser.discardBrowser(win.gBrowser.tabs[1]);
            """
            )

        window_handles = self.marionette.window_handles
        window_handles.remove(self.start_tab)
        self.assertEqual(1, len(window_handles))
        self.marionette.switch_to_window(window_handles[0], focus=False)
        self.marionette.close()
        self.assertListEqual([self.start_tab], self.marionette.window_handles)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import os

from setuptools import setup

PACKAGE_VERSION = "1.0.1"

try:
    pwd = os.path.dirname(os.path.abspath(__file__))
    description = open(os.path.join(pwd, "README.rst")).read()
except Exception:
    description = ""

setup(
    name="mozsystemmonitor",
    description="Monitor system resource usage.",
    long_description="see https://firefox-source-docs.mozilla.org/mozbase/index.html",
    classifiers=[
        "Programming Language :: Python :: 3",
    ],
    # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
    license="MPL 2.0",
    keywords="mozilla",
    author="Mozilla Automation and Tools Team",
    author_email="tools@lists.mozilla.org",
    url="https://wiki.mozilla.org/Auto-tools/Projects/Mozbase",
    packages=["mozsystemmonitor"],
    version=PACKAGE_VERSION,
    install_requires=["psutil >= 3.1.1"],
)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

from setuptools import setup

PACKAGE_NAME = "mozfile"
PACKAGE_VERSION = "3.0.0"

setup(
    name=PACKAGE_NAME,
    version=PACKAGE_VERSION,
    description="Library of file utilities for use in Mozilla testing",
    long_description="see https://firefox-source-docs.mozilla.org/mozbase/index.html",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
    ],
    keywords="mozilla",
    author="Mozilla Automation and Tools team",
    author_email="tools@lists.mozilla.org",
    url="https://wiki.mozilla.org/Auto-tools/Projects/Mozbase",
    license="MPL",
    packages=["mozfile"],
    include_package_data=True,
    zip_safe=False,
    install_requires=["six >= 1.13.0"],
    tests_require=["wptserve"],
)



import pluggy

hookspec = pluggy.HookspecMarker("myproject")
hookimpl = pluggy.HookimplMarker("myproject")


class MySpec:
    """A hook specification namespace."""

    @hookspec
    def myhook(self, arg1, arg2):
        """My special little hook that you can customize."""


class Plugin_1:
    """A hook implementation namespace."""

    @hookimpl
    def myhook(self, arg1, arg2):
        print("inside Plugin_1.myhook()")
        return arg1 + arg2


class Plugin_2:
    """A 2nd hook implementation namespace."""

    @hookimpl
    def myhook(self, arg1, arg2):
        print("inside Plugin_2.myhook()")
        return arg1 - arg2


# create a manager and add the spec
pm = pluggy.PluginManager("myproject")
pm.add_hookspecs(MySpec)
# register plugins
pm.register(Plugin_1())
pm.register(Plugin_2())
# call our `myhook` hook
results = pm.hook.myhook(arg1=1, arg2=2)
print(results)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import json
import os
import shutil
from datetime import datetime
from shutil import copyfile
from zipfile import ZipFile

from .utils import LOG


class RecordingFile:
    def __init__(self, path_to_zip_file):
        self._recording_zip_path = path_to_zip_file

        self._base_name = os.path.splitext(os.path.basename(self._recording_zip_path))[
            0
        ]
        if not os.path.splitext(path_to_zip_file)[1] == ".zip":
            LOG.error(
                "Wrong file type! The provided recording should be a zip file. %s"
                % path_to_zip_file
            )
            raise Exception(
                "Wrong file type! The provided recording should be a zip file."
            )

        # create a temp dir
        self._mozproxy_dir = os.environ["MOZPROXY_DIR"]
        self._tempdir = os.path.join(self._mozproxy_dir, self._base_name)

        if os.path.exists(self._tempdir):
            LOG.info(
                "The recording dir already exists! Resetting the existing dir and data."
            )
            shutil.rmtree(self._tempdir)
        os.mkdir(self._tempdir)

        self._metadata_path = self._get_temp_path("metadata.json")
        self._recording = self._get_temp_path("dump.mp")

        if os.path.exists(path_to_zip_file):
            with ZipFile(path_to_zip_file, "r") as zipObj:
                # Extract all the contents of zip file in different directory
                zipObj.extractall(self._tempdir)

            if not os.path.exists(self._recording):
                self._convert_to_new_format()

            if not os.path.exists(self._metadata_path):
                LOG.error("metadata file is missing!")
                raise Exception("metadata file is missing!")

            with open(self._metadata_path) as json_file:
                self._metadata = json.load(json_file)
            self.validate_recording()
            LOG.info(
                "Loaded recoording generated on %s" % self.metadata("recording_date")
            )
        else:
            LOG.info("Recording file does not exists!!! Generating base structure")
            self._metadata = {"content": [], "recording_date": str(datetime.now())}

    def _convert_to_new_format(self):
        # Convert zip recording to new format

        LOG.info("Convert zip recording to new format")

        for tmp_file in os.listdir(self._tempdir):
            if tmp_file.endswith(".mp"):
                LOG.info("Renaming %s to dump.mp file" % tmp_file)
                os.rename(self._get_temp_path(tmp_file), self._get_temp_path("dump.mp"))
            elif tmp_file.endswith(".json"):
                if tmp_file.startswith("mitm_netlocs_"):
                    LOG.info("Renaming %s to netlocs.json file" % tmp_file)
                    os.rename(
                        self._get_temp_path("%s.json" % os.path.splitext(tmp_file)[0]),
                        self._get_temp_path("netlocs.json"),
                    )
                else:
                    LOG.info("Renaming %s to metadata.json file" % tmp_file)
                    os.rename(
                        self._get_temp_path("%s.json" % os.path.splitext(tmp_file)[0]),
                        self._get_temp_path("metadata.json"),
                    )
            elif tmp_file.endswith(".png"):
                LOG.info("Renaming %s to screenshot.png file" % tmp_file)
                os.rename(
                    self._get_temp_path("%s.png" % os.path.splitext(tmp_file)[0]),
                    self._get_temp_path("screenshot.png"),
                )

    def _get_temp_path(self, file_name):
        return os.path.join(self._tempdir, file_name)

    def validate_recording(self):
        # Validates that minimum zip file content exists
        if not os.path.exists(self._recording):
            LOG.error("Recording file is missing!")
            raise Exception("Recording file is missing!")

        if not os.path.exists(self._metadata_path):
            LOG.error("Metadata file is missing!")
            raise Exception("Metadata file is missing!")

        if "content" in self._metadata:
            # check that all extra files specified in the recording are present
            for content_file in self._metadata["content"]:
                if not os.path.exists(self._get_temp_path(content_file)):
                    LOG.error("File %s does not exist!!" % content_file)
                    raise Exception("Recording file is missing!")
        else:
            LOG.info("Old file type! Not confirming content!")

    def metadata(self, name):
        # Return metadata value
        return self._metadata[name]

    def set_metadata(self, entry, value):
        # Set metadata value
        self._metadata[entry] = value

    @property
    def recording_path(self):
        # Returns the path of the recoring.mp file included in the zip
        return self._recording

    def get_file(self, file_name):
        # Returns the path to a specified file included in the recording zip
        return self._get_temp_path(file_name)

    def add_file(self, path):
        # Adds file to Zip
        if os.path.exists(path):
            copyfile(path, self._tempdir)
            self._metadata["content"].append(os.path.basename(path))
        else:
            LOG.error("Target file %s does not exist!!" % path)
            raise Exception("File does not exist!!!")

    def update_metadata(self):
        # Update metadata with information generated by HttpProtocolExtractor plugin
        # This data is geerated after closing the MitMPtoxy process

        dump_file = self._get_temp_path("dump.json")
        if os.path.exists(dump_file):
            with open(dump_file) as dump:
                dump_info = json.load(dump)
                self._metadata.update(dump_info)

    def generate_zip_file(self):
        self.update_metadata()
        with open(self._get_temp_path(self._metadata_path), "w") as metadata_file:
            json.dump(self._metadata, metadata_file, sort_keys=True, indent=4)

        with ZipFile(self._recording_zip_path, "w") as zf:
            zf.write(self._metadata_path, "metadata.json")
            zf.write(self.recording_path, "dump.mp")
            for file in self._metadata["content"]:
                zf.write(self._get_temp_path(file), file)
        LOG.info("Generated new recording file at: %s" % self._recording_zip_path)



import unicodedata
from functools import lru_cache


@lru_cache(100)
def wcwidth(c: str) -> int:
    """Determine how many columns are needed to display a character in a terminal.

    Returns -1 if the character is not printable.
    Returns 0, 1 or 2 for other characters.
    """
    o = ord(c)

    # ASCII fast path.
    if 0x20 <= o < 0x07F:
        return 1

    # Some Cf/Zp/Zl characters which should be zero-width.
    if (
        o == 0x0000
        or 0x200B <= o <= 0x200F
        or 0x2028 <= o <= 0x202E
        or 0x2060 <= o <= 0x2063
    ):
        return 0

    category = unicodedata.category(c)

    # Control characters.
    if category == "Cc":
        return -1

    # Combining characters with zero width.
    if category in ("Me", "Mn"):
        return 0

    # Full/Wide east asian characters.
    if unicodedata.east_asian_width(c) in ("F", "W"):
        return 2

    return 1


def wcswidth(s: str) -> int:
    """Determine how many columns are needed to display a string in a terminal.

    Returns -1 if the string contains non-printable characters.
    """
    width = 0
    for c in unicodedata.normalize("NFC", s):
        wc = wcwidth(c)
        if wc < 0:
            return -1
        width += wc
    return width



# Make flake8, ruff and autodoc happy
if "self" not in globals():
    self = None

# This code is loaded via `mach python --exec-file`, so it runs in the scope of
# the `mach python` command.
old = self._mach_context.post_dispatch_handler


def handler(context, handler, instance, result, start_time, end_time, depth, args):
    global old
    # Round off sub-second precision.
    old(context, handler, instance, result, int(start_time), end_time, depth, args)


self._mach_context.post_dispatch_handler = handler



# -*- coding: utf-8 -*-
import pytest
import os
import json
import sys

from hypothesis.strategies import text

if sys.version_info[0] == 2:
    from codecs import open

# We need to grab one text example from hypothesis to prime its cache.
text().example()

# This pair of generator expressions are pretty lame, but building lists is a
# bad idea as I plan to have a substantial number of tests here.
story_directories = (
    os.path.join('test/test_fixtures', d)
    for d in os.listdir('test/test_fixtures')
)
story_files = (
    os.path.join(storydir, name)
    for storydir in story_directories
    for name in os.listdir(storydir)
    if 'raw-data' not in storydir
)
raw_story_files = (
    os.path.join('test/test_fixtures/raw-data', name)
    for name in os.listdir('test/test_fixtures/raw-data')
)


@pytest.fixture(scope='class', params=story_files)
def story(request):
    """
    Provides a detailed HPACK story to test with.
    """
    with open(request.param, 'r', encoding='utf-8') as f:
        return json.load(f)


@pytest.fixture(scope='class', params=raw_story_files)
def raw_story(request):
    """
    Provides a detailed HPACK story to test the encoder with.
    """
    with open(request.param, 'r', encoding='utf-8') as f:
        return json.load(f)



#!/usr/bin/env python

import mozunit


def test_escape_command_line(mock_adb_object, redirect_stdout_and_assert):
    """Test _escape_command_line."""
    cases = {
        # expected output : test input
        "adb shell ls -l": ["adb", "shell", "ls", "-l"],
        "adb shell 'ls -l'": ["adb", "shell", "ls -l"],
        "-e 'if (true)'": ["-e", "if (true)"],
        "-e 'if (x === \"hello\")'": ["-e", 'if (x === "hello")'],
        "-e 'if (x === '\"'\"'hello'\"'\"')'": ["-e", "if (x === 'hello')"],
    }
    for expected, input in cases.items():
        assert mock_adb_object._escape_command_line(input) == expected


if __name__ == "__main__":
    mozunit.main()



from typing import ClassVar, Generic, Iterable, Text, Type, Union
from typing_extensions import Final

class raw:
    uniobj: Final[Text]
    def __init__(self, uniobj: Text) -> None: ...

class _NamespaceMetaclass(type):
    def __getattr__(self, name: str) -> Type[Tag]: ...

class Namespace(metaclass=_NamespaceMetaclass): ...

class Tag(list):
    class Attr:
        def __getattr__(self, attr: str) -> Text: ...
    attr: Final[Attr]
    def __init__(self, *args: Union[Text, raw, Tag, Iterable[Tag]], **kwargs: Union[Text, raw]) -> None: ...
    def unicode(self, indent: int = ...) -> Text: ...

class html(Namespace):
    class Style:
        def __init__(self, **kw: Union[str, Text]) -> None: ...
    style: ClassVar[Style]

def escape(ustring: Union[str, Text]) -> Text: ...



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from marionette_harness import MarionetteTestCase

EXT_ID = "extension-with-bg-sw@test"
EXT_DIR_PATH = "extension-with-bg-sw"
PREF_BG_SW_ENABLED = "extensions.backgroundServiceWorker.enabled"
PREF_PERSIST_TEMP_ADDONS = (
    "dom.serviceWorkers.testing.persistTemporarilyInstalledAddons"
)


class MarionetteServiceWorkerTestCase(MarionetteTestCase):
    def get_extension_url(self, path="/"):
        with self.marionette.using_context("chrome"):
            return self.marionette.execute_script(
                """
              let policy = WebExtensionPolicy.getByID(arguments[0]);
              return policy.getURL(arguments[1])
            """,
                script_args=(self.test_extension_id, path),
            )

    @property
    def is_extension_service_worker_registered(self):
        with self.marionette.using_context("chrome"):
            return self.marionette.execute_script(
                """
                let serviceWorkerManager = Cc["@mozilla.org/serviceworkers/manager;1"].getService(
                    Ci.nsIServiceWorkerManager
                );

                let serviceWorkers = serviceWorkerManager.getAllRegistrations();
                for (let i = 0; i < serviceWorkers.length; i++) {
                    let sw = serviceWorkers.queryElementAt(
                        i,
                        Ci.nsIServiceWorkerRegistrationInfo
                    );
                    if (sw.scope == arguments[0]) {
                        return true;
                    }
                }
                return false;
            """,
                script_args=(self.test_extension_base_url,),
            )



import sys

if sys.version_info >= (3, 8):
    from importlib import metadata
else:
    import importlib_metadata as metadata


extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.coverage",
    "sphinx.ext.viewcode",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.

project = "pluggy"
copyright = "2016, Holger Krekel"
author = "Holger Krekel"

release = metadata.version(project)
# The short X.Y version.
version = ".".join(release.split(".")[:2])


language = None

pygments_style = "sphinx"
# html_logo = "_static/img/plug.png"
html_theme = "alabaster"
html_theme_options = {
    "logo": "img/plug.png",
    "description": "The pytest plugin system",
    "github_user": "pytest-dev",
    "github_repo": "pluggy",
    "github_button": "true",
    "github_banner": "true",
    "github_type": "star",
    "badge_branch": "master",
    "page_width": "1080px",
    "fixed_sidebar": "false",
}
html_sidebars = {
    "**": ["about.html", "localtoc.html", "relations.html", "searchbox.html"]
}
html_static_path = ["_static"]

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "pluggy", "pluggy Documentation", [author], 1)]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "pluggy",
        "pluggy Documentation",
        author,
        "pluggy",
        "One line description of project.",
        "Miscellaneous",
    )
]

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "pytest": ("https://docs.pytest.org/en/latest", None),
    "setuptools": ("https://setuptools.readthedocs.io/en/latest", None),
    "tox": ("https://tox.readthedocs.io/en/latest", None),
    "devpi": ("https://devpi.net/docs/devpi/devpi/stable/+doc/", None),
}



# -*- coding: utf-8 -*-
"""
head_request.py
~~~~~~~~~~~~~~~

A short example that demonstrates a client that makes HEAD requests to certain
websites.

This example is intended as a reproduction of nghttp2 issue 396, for the
purposes of compatibility testing.
"""
from __future__ import print_function

from twisted.internet import reactor
from twisted.internet.endpoints import connectProtocol, SSL4ClientEndpoint
from twisted.internet.protocol import Protocol
from twisted.internet.ssl import optionsForClientTLS
from hyperframe.frame import SettingsFrame
from h2.connection import H2Connection
from h2.events import (
    ResponseReceived, DataReceived, StreamEnded,
    StreamReset, SettingsAcknowledged,
)


AUTHORITY = u'nghttp2.org'
PATH = '/httpbin/'
SIZE = 4096


class H2Protocol(Protocol):
    def __init__(self):
        self.conn = H2Connection()
        self.known_proto = None
        self.request_made = False

    def connectionMade(self):
        self.conn.initiate_connection()

        # This reproduces the error in #396, by changing the header table size.
        self.conn.update_settings({SettingsFrame.HEADER_TABLE_SIZE: SIZE})

        self.transport.write(self.conn.data_to_send())

    def dataReceived(self, data):
        if not self.known_proto:
            self.known_proto = self.transport.negotiatedProtocol
            assert self.known_proto == b'h2'

        events = self.conn.receive_data(data)

        for event in events:
            if isinstance(event, ResponseReceived):
                self.handleResponse(event.headers, event.stream_id)
            elif isinstance(event, DataReceived):
                self.handleData(event.data, event.stream_id)
            elif isinstance(event, StreamEnded):
                self.endStream(event.stream_id)
            elif isinstance(event, SettingsAcknowledged):
                self.settingsAcked(event)
            elif isinstance(event, StreamReset):
                reactor.stop()
                raise RuntimeError("Stream reset: %d" % event.error_code)
            else:
                print(event)

        data = self.conn.data_to_send()
        if data:
            self.transport.write(data)

    def settingsAcked(self, event):
        # Having received the remote settings change, lets send our request.
        if not self.request_made:
            self.sendRequest()

    def handleResponse(self, response_headers, stream_id):
        for name, value in response_headers:
            print("%s: %s" % (name.decode('utf-8'), value.decode('utf-8')))

        print("")

    def handleData(self, data, stream_id):
        print(data, end='')

    def endStream(self, stream_id):
        self.conn.close_connection()
        self.transport.write(self.conn.data_to_send())
        self.transport.loseConnection()
        reactor.stop()

    def sendRequest(self):
        request_headers = [
            (':method', 'HEAD'),
            (':authority', AUTHORITY),
            (':scheme', 'https'),
            (':path', PATH),
        ]
        self.conn.send_headers(1, request_headers, end_stream=True)
        self.request_made = True

options = optionsForClientTLS(
    hostname=AUTHORITY,
    acceptableProtocols=[b'h2'],
)

connectProtocol(
    SSL4ClientEndpoint(reactor, AUTHORITY, 443, options),
    H2Protocol()
)
reactor.run()



#!/usr/bin/env python

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import mozunit
import pytest
from mozprofile.permissions import (
    BadPortLocationError,
    LocationsSyntaxError,
    MissingPrimaryLocationError,
    MultiplePrimaryLocationsError,
    ServerLocations,
)

LOCATIONS = """# This is the primary location from which tests run.
#
http://mochi.test:8888          primary,privileged

# a few test locations
http://127.0.0.1:80             privileged
http://127.0.0.1:8888           privileged
https://test:80                 privileged
http://example.org:80           privileged
http://test1.example.org        privileged

"""

LOCATIONS_NO_PRIMARY = """http://secondary.test:80        privileged
http://tertiary.test:8888       privileged
"""

LOCATIONS_BAD_PORT = """http://mochi.test:8888  primary,privileged
http://127.0.0.1:80             privileged
http://127.0.0.1:8888           privileged
http://test:badport             privileged
http://example.org:80           privileged
"""


def compare_location(location, scheme, host, port, options):
    assert location.scheme == scheme
    assert location.host == host
    assert location.port == port
    assert location.options == options


@pytest.fixture
def create_temp_file(tmpdir):
    def inner(contents):
        f = tmpdir.mkdtemp().join("locations.txt")
        f.write(contents)
        return f.strpath

    return inner


def test_server_locations(create_temp_file):
    # write a permissions file
    f = create_temp_file(LOCATIONS)

    # read the locations
    locations = ServerLocations(f)

    # ensure that they're what we expect
    assert len(locations) == 6
    i = iter(locations)
    compare_location(next(i), "http", "mochi.test", "8888", ["primary", "privileged"])
    compare_location(next(i), "http", "127.0.0.1", "80", ["privileged"])
    compare_location(next(i), "http", "127.0.0.1", "8888", ["privileged"])
    compare_location(next(i), "https", "test", "80", ["privileged"])
    compare_location(next(i), "http", "example.org", "80", ["privileged"])
    compare_location(next(i), "http", "test1.example.org", "8888", ["privileged"])

    locations.add_host("mozilla.org")
    assert len(locations) == 7
    compare_location(next(i), "http", "mozilla.org", "80", ["privileged"])

    # test some errors
    with pytest.raises(MultiplePrimaryLocationsError):
        locations.add_host("primary.test", options="primary")

    # assert we don't throw DuplicateLocationError
    locations.add_host("127.0.0.1")

    with pytest.raises(BadPortLocationError):
        locations.add_host("127.0.0.1", port="abc")

    # test some errors in locations file
    f = create_temp_file(LOCATIONS_NO_PRIMARY)

    exc = None
    try:
        ServerLocations(f)
    except LocationsSyntaxError as e:
        exc = e
    assert exc is not None
    assert exc.err.__class__ == MissingPrimaryLocationError
    assert exc.lineno == 3

    # test bad port in a locations file to ensure lineno calculated
    # properly.
    f = create_temp_file(LOCATIONS_BAD_PORT)

    exc = None
    try:
        ServerLocations(f)
    except LocationsSyntaxError as e:
        exc = e
    assert exc is not None
    assert exc.err.__class__ == BadPortLocationError
    assert exc.lineno == 4


if __name__ == "__main__":
    mozunit.main()



import re
import warnings
from pathlib import Path
from typing import cast
from typing import List
from typing import Type

import pytest
from _pytest import nodes
from _pytest.compat import legacy_path
from _pytest.outcomes import OutcomeException
from _pytest.pytester import Pytester
from _pytest.warning_types import PytestWarning


@pytest.mark.parametrize(
    ("nodeid", "expected"),
    (
        ("", [""]),
        ("a", ["", "a"]),
        ("aa/b", ["", "aa", "aa/b"]),
        ("a/b/c", ["", "a", "a/b", "a/b/c"]),
        ("a/bbb/c::D", ["", "a", "a/bbb", "a/bbb/c", "a/bbb/c::D"]),
        ("a/b/c::D::eee", ["", "a", "a/b", "a/b/c", "a/b/c::D", "a/b/c::D::eee"]),
        ("::xx", ["", "::xx"]),
        # / only considered until first ::
        ("a/b/c::D/d::e", ["", "a", "a/b", "a/b/c", "a/b/c::D/d", "a/b/c::D/d::e"]),
        # : alone is not a separator.
        ("a/b::D:e:f::g", ["", "a", "a/b", "a/b::D:e:f", "a/b::D:e:f::g"]),
        # / not considered if a part of a test name
        ("a/b::c/d::e[/test]", ["", "a", "a/b", "a/b::c/d", "a/b::c/d::e[/test]"]),
    ),
)
def test_iterparentnodeids(nodeid: str, expected: List[str]) -> None:
    result = list(nodes.iterparentnodeids(nodeid))
    assert result == expected


def test_node_from_parent_disallowed_arguments() -> None:
    with pytest.raises(TypeError, match="session is"):
        nodes.Node.from_parent(None, session=None)  # type: ignore[arg-type]
    with pytest.raises(TypeError, match="config is"):
        nodes.Node.from_parent(None, config=None)  # type: ignore[arg-type]


def test_node_direct_construction_deprecated() -> None:
    with pytest.raises(
        OutcomeException,
        match=(
            "Direct construction of _pytest.nodes.Node has been deprecated, please "
            "use _pytest.nodes.Node.from_parent.\nSee "
            "https://docs.pytest.org/en/stable/deprecations.html#node-construction-changed-to-node-from-parent"
            " for more details."
        ),
    ):
        nodes.Node(None, session=None)  # type: ignore[arg-type]


def test_subclassing_both_item_and_collector_deprecated(
    request, tmp_path: Path
) -> None:
    """
    Verifies we warn on diamond inheritance as well as correctly managing legacy
    inheritance constructors with missing args as found in plugins.
    """

    # We do not expect any warnings messages to issued during class definition.
    with warnings.catch_warnings():
        warnings.simplefilter("error")

        class SoWrong(nodes.Item, nodes.File):
            def __init__(self, fspath, parent):
                """Legacy ctor with legacy call # don't wana see"""
                super().__init__(fspath, parent)

    with pytest.warns(PytestWarning) as rec:
        SoWrong.from_parent(
            request.session, fspath=legacy_path(tmp_path / "broken.txt")
        )
    messages = [str(x.message) for x in rec]
    assert any(
        re.search(".*SoWrong.* not using a cooperative constructor.*", x)
        for x in messages
    )
    assert any(
        re.search("(?m)SoWrong .* should not be a collector", x) for x in messages
    )


@pytest.mark.parametrize(
    "warn_type, msg", [(DeprecationWarning, "deprecated"), (PytestWarning, "pytest")]
)
def test_node_warn_is_no_longer_only_pytest_warnings(
    pytester: Pytester, warn_type: Type[Warning], msg: str
) -> None:
    items = pytester.getitems(
        """
        def test():
            pass
    """
    )
    with pytest.warns(warn_type, match=msg):
        items[0].warn(warn_type(msg))


def test_node_warning_enforces_warning_types(pytester: Pytester) -> None:
    items = pytester.getitems(
        """
        def test():
            pass
    """
    )
    with pytest.raises(
        ValueError, match="warning must be an instance of Warning or subclass"
    ):
        items[0].warn(Exception("ok"))  # type: ignore[arg-type]


def test__check_initialpaths_for_relpath() -> None:
    """Ensure that it handles dirs, and does not always use dirname."""
    cwd = Path.cwd()

    class FakeSession1:
        _initialpaths = frozenset({cwd})

    session = cast(pytest.Session, FakeSession1)

    assert nodes._check_initialpaths_for_relpath(session, cwd) == ""

    sub = cwd / "file"

    class FakeSession2:
        _initialpaths = frozenset({cwd})

    session = cast(pytest.Session, FakeSession2)

    assert nodes._check_initialpaths_for_relpath(session, sub) == "file"

    outside = Path("/outside-this-does-not-exist")
    assert nodes._check_initialpaths_for_relpath(session, outside) is None


def test_failure_with_changed_cwd(pytester: Pytester) -> None:
    """
    Test failure lines should use absolute paths if cwd has changed since
    invocation, so the path is correct (#6428).
    """
    p = pytester.makepyfile(
        """
        import os
        import pytest

        @pytest.fixture
        def private_dir():
            out_dir = 'ddd'
            os.mkdir(out_dir)
            old_dir = os.getcwd()
            os.chdir(out_dir)
            yield out_dir
            os.chdir(old_dir)

        def test_show_wrong_path(private_dir):
            assert False
    """
    )
    result = pytester.runpytest()
    result.stdout.fnmatch_lines([str(p) + ":*: AssertionError", "*1 failed in *"])



"""Discover and run std-library "unittest" style tests."""
import sys
import traceback
import types
from typing import Any
from typing import Callable
from typing import Generator
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import TYPE_CHECKING
from typing import Union

import _pytest._code
import pytest
from _pytest.compat import getimfunc
from _pytest.compat import is_async_function
from _pytest.config import hookimpl
from _pytest.fixtures import FixtureRequest
from _pytest.nodes import Collector
from _pytest.nodes import Item
from _pytest.outcomes import exit
from _pytest.outcomes import fail
from _pytest.outcomes import skip
from _pytest.outcomes import xfail
from _pytest.python import Class
from _pytest.python import Function
from _pytest.python import PyCollector
from _pytest.runner import CallInfo
from _pytest.scope import Scope

if TYPE_CHECKING:
    import unittest
    import twisted.trial.unittest

    _SysExcInfoType = Union[
        Tuple[Type[BaseException], BaseException, types.TracebackType],
        Tuple[None, None, None],
    ]


def pytest_pycollect_makeitem(
    collector: PyCollector, name: str, obj: object
) -> Optional["UnitTestCase"]:
    # Has unittest been imported and is obj a subclass of its TestCase?
    try:
        ut = sys.modules["unittest"]
        # Type ignored because `ut` is an opaque module.
        if not issubclass(obj, ut.TestCase):  # type: ignore
            return None
    except Exception:
        return None
    # Yes, so let's collect it.
    item: UnitTestCase = UnitTestCase.from_parent(collector, name=name, obj=obj)
    return item


class UnitTestCase(Class):
    # Marker for fixturemanger.getfixtureinfo()
    # to declare that our children do not support funcargs.
    nofuncargs = True

    def collect(self) -> Iterable[Union[Item, Collector]]:
        from unittest import TestLoader

        cls = self.obj
        if not getattr(cls, "__test__", True):
            return

        skipped = _is_skipped(cls)
        if not skipped:
            self._inject_setup_teardown_fixtures(cls)
            self._inject_setup_class_fixture()

        self.session._fixturemanager.parsefactories(self, unittest=True)
        loader = TestLoader()
        foundsomething = False
        for name in loader.getTestCaseNames(self.obj):
            x = getattr(self.obj, name)
            if not getattr(x, "__test__", True):
                continue
            funcobj = getimfunc(x)
            yield TestCaseFunction.from_parent(self, name=name, callobj=funcobj)
            foundsomething = True

        if not foundsomething:
            runtest = getattr(self.obj, "runTest", None)
            if runtest is not None:
                ut = sys.modules.get("twisted.trial.unittest", None)
                # Type ignored because `ut` is an opaque module.
                if ut is None or runtest != ut.TestCase.runTest:  # type: ignore
                    yield TestCaseFunction.from_parent(self, name="runTest")

    def _inject_setup_teardown_fixtures(self, cls: type) -> None:
        """Injects a hidden auto-use fixture to invoke setUpClass/setup_method and corresponding
        teardown functions (#517)."""
        class_fixture = _make_xunit_fixture(
            cls,
            "setUpClass",
            "tearDownClass",
            "doClassCleanups",
            scope=Scope.Class,
            pass_self=False,
        )
        if class_fixture:
            cls.__pytest_class_setup = class_fixture  # type: ignore[attr-defined]

        method_fixture = _make_xunit_fixture(
            cls,
            "setup_method",
            "teardown_method",
            None,
            scope=Scope.Function,
            pass_self=True,
        )
        if method_fixture:
            cls.__pytest_method_setup = method_fixture  # type: ignore[attr-defined]


def _make_xunit_fixture(
    obj: type,
    setup_name: str,
    teardown_name: str,
    cleanup_name: Optional[str],
    scope: Scope,
    pass_self: bool,
):
    setup = getattr(obj, setup_name, None)
    teardown = getattr(obj, teardown_name, None)
    if setup is None and teardown is None:
        return None

    if cleanup_name:
        cleanup = getattr(obj, cleanup_name, lambda *args: None)
    else:

        def cleanup(*args):
            pass

    @pytest.fixture(
        scope=scope.value,
        autouse=True,
        # Use a unique name to speed up lookup.
        name=f"_unittest_{setup_name}_fixture_{obj.__qualname__}",
    )
    def fixture(self, request: FixtureRequest) -> Generator[None, None, None]:
        if _is_skipped(self):
            reason = self.__unittest_skip_why__
            raise pytest.skip.Exception(reason, _use_item_location=True)
        if setup is not None:
            try:
                if pass_self:
                    setup(self, request.function)
                else:
                    setup()
            # unittest does not call the cleanup function for every BaseException, so we
            # follow this here.
            except Exception:
                if pass_self:
                    cleanup(self)
                else:
                    cleanup()

                raise
        yield
        try:
            if teardown is not None:
                if pass_self:
                    teardown(self, request.function)
                else:
                    teardown()
        finally:
            if pass_self:
                cleanup(self)
            else:
                cleanup()

    return fixture


class TestCaseFunction(Function):
    nofuncargs = True
    _excinfo: Optional[List[_pytest._code.ExceptionInfo[BaseException]]] = None
    _testcase: Optional["unittest.TestCase"] = None

    def _getobj(self):
        assert self.parent is not None
        # Unlike a regular Function in a Class, where `item.obj` returns
        # a *bound* method (attached to an instance), TestCaseFunction's
        # `obj` returns an *unbound* method (not attached to an instance).
        # This inconsistency is probably not desirable, but needs some
        # consideration before changing.
        return getattr(self.parent.obj, self.originalname)  # type: ignore[attr-defined]

    def setup(self) -> None:
        # A bound method to be called during teardown() if set (see 'runtest()').
        self._explicit_tearDown: Optional[Callable[[], None]] = None
        assert self.parent is not None
        self._testcase = self.parent.obj(self.name)  # type: ignore[attr-defined]
        self._obj = getattr(self._testcase, self.name)
        if hasattr(self, "_request"):
            self._request._fillfixtures()

    def teardown(self) -> None:
        if self._explicit_tearDown is not None:
            self._explicit_tearDown()
            self._explicit_tearDown = None
        self._testcase = None
        self._obj = None

    def startTest(self, testcase: "unittest.TestCase") -> None:
        pass

    def _addexcinfo(self, rawexcinfo: "_SysExcInfoType") -> None:
        # Unwrap potential exception info (see twisted trial support below).
        rawexcinfo = getattr(rawexcinfo, "_rawexcinfo", rawexcinfo)
        try:
            excinfo = _pytest._code.ExceptionInfo[BaseException].from_exc_info(rawexcinfo)  # type: ignore[arg-type]
            # Invoke the attributes to trigger storing the traceback
            # trial causes some issue there.
            excinfo.value
            excinfo.traceback
        except TypeError:
            try:
                try:
                    values = traceback.format_exception(*rawexcinfo)
                    values.insert(
                        0,
                        "NOTE: Incompatible Exception Representation, "
                        "displaying natively:\n\n",
                    )
                    fail("".join(values), pytrace=False)
                except (fail.Exception, KeyboardInterrupt):
                    raise
                except BaseException:
                    fail(
                        "ERROR: Unknown Incompatible Exception "
                        "representation:\n%r" % (rawexcinfo,),
                        pytrace=False,
                    )
            except KeyboardInterrupt:
                raise
            except fail.Exception:
                excinfo = _pytest._code.ExceptionInfo.from_current()
        self.__dict__.setdefault("_excinfo", []).append(excinfo)

    def addError(
        self, testcase: "unittest.TestCase", rawexcinfo: "_SysExcInfoType"
    ) -> None:
        try:
            if isinstance(rawexcinfo[1], exit.Exception):
                exit(rawexcinfo[1].msg)
        except TypeError:
            pass
        self._addexcinfo(rawexcinfo)

    def addFailure(
        self, testcase: "unittest.TestCase", rawexcinfo: "_SysExcInfoType"
    ) -> None:
        self._addexcinfo(rawexcinfo)

    def addSkip(self, testcase: "unittest.TestCase", reason: str) -> None:
        try:
            raise pytest.skip.Exception(reason, _use_item_location=True)
        except skip.Exception:
            self._addexcinfo(sys.exc_info())

    def addExpectedFailure(
        self,
        testcase: "unittest.TestCase",
        rawexcinfo: "_SysExcInfoType",
        reason: str = "",
    ) -> None:
        try:
            xfail(str(reason))
        except xfail.Exception:
            self._addexcinfo(sys.exc_info())

    def addUnexpectedSuccess(
        self,
        testcase: "unittest.TestCase",
        reason: Optional["twisted.trial.unittest.Todo"] = None,
    ) -> None:
        msg = "Unexpected success"
        if reason:
            msg += f": {reason.reason}"
        # Preserve unittest behaviour - fail the test. Explicitly not an XPASS.
        try:
            fail(msg, pytrace=False)
        except fail.Exception:
            self._addexcinfo(sys.exc_info())

    def addSuccess(self, testcase: "unittest.TestCase") -> None:
        pass

    def stopTest(self, testcase: "unittest.TestCase") -> None:
        pass

    def runtest(self) -> None:
        from _pytest.debugging import maybe_wrap_pytest_function_for_tracing

        assert self._testcase is not None

        maybe_wrap_pytest_function_for_tracing(self)

        # Let the unittest framework handle async functions.
        if is_async_function(self.obj):
            # Type ignored because self acts as the TestResult, but is not actually one.
            self._testcase(result=self)  # type: ignore[arg-type]
        else:
            # When --pdb is given, we want to postpone calling tearDown() otherwise
            # when entering the pdb prompt, tearDown() would have probably cleaned up
            # instance variables, which makes it difficult to debug.
            # Arguably we could always postpone tearDown(), but this changes the moment where the
            # TestCase instance interacts with the results object, so better to only do it
            # when absolutely needed.
            if self.config.getoption("usepdb") and not _is_skipped(self.obj):
                self._explicit_tearDown = self._testcase.tearDown
                setattr(self._testcase, "tearDown", lambda *args: None)

            # We need to update the actual bound method with self.obj, because
            # wrap_pytest_function_for_tracing replaces self.obj by a wrapper.
            setattr(self._testcase, self.name, self.obj)
            try:
                self._testcase(result=self)  # type: ignore[arg-type]
            finally:
                delattr(self._testcase, self.name)

    def _prunetraceback(
        self, excinfo: _pytest._code.ExceptionInfo[BaseException]
    ) -> None:
        super()._prunetraceback(excinfo)
        traceback = excinfo.traceback.filter(
            lambda x: not x.frame.f_globals.get("__unittest")
        )
        if traceback:
            excinfo.traceback = traceback


@hookimpl(tryfirst=True)
def pytest_runtest_makereport(item: Item, call: CallInfo[None]) -> None:
    if isinstance(item, TestCaseFunction):
        if item._excinfo:
            call.excinfo = item._excinfo.pop(0)
            try:
                del call.result
            except AttributeError:
                pass

    # Convert unittest.SkipTest to pytest.skip.
    # This is actually only needed for nose, which reuses unittest.SkipTest for
    # its own nose.SkipTest. For unittest TestCases, SkipTest is already
    # handled internally, and doesn't reach here.
    unittest = sys.modules.get("unittest")
    if (
        unittest
        and call.excinfo
        and isinstance(call.excinfo.value, unittest.SkipTest)  # type: ignore[attr-defined]
    ):
        excinfo = call.excinfo
        call2 = CallInfo[None].from_call(
            lambda: pytest.skip(str(excinfo.value)), call.when
        )
        call.excinfo = call2.excinfo


# Twisted trial support.


@hookimpl(hookwrapper=True)
def pytest_runtest_protocol(item: Item) -> Generator[None, None, None]:
    if isinstance(item, TestCaseFunction) and "twisted.trial.unittest" in sys.modules:
        ut: Any = sys.modules["twisted.python.failure"]
        Failure__init__ = ut.Failure.__init__
        check_testcase_implements_trial_reporter()

        def excstore(
            self, exc_value=None, exc_type=None, exc_tb=None, captureVars=None
        ):
            if exc_value is None:
                self._rawexcinfo = sys.exc_info()
            else:
                if exc_type is None:
                    exc_type = type(exc_value)
                self._rawexcinfo = (exc_type, exc_value, exc_tb)
            try:
                Failure__init__(
                    self, exc_value, exc_type, exc_tb, captureVars=captureVars
                )
            except TypeError:
                Failure__init__(self, exc_value, exc_type, exc_tb)

        ut.Failure.__init__ = excstore
        yield
        ut.Failure.__init__ = Failure__init__
    else:
        yield


def check_testcase_implements_trial_reporter(done: List[int] = []) -> None:
    if done:
        return
    from zope.interface import classImplements
    from twisted.trial.itrial import IReporter

    classImplements(TestCaseFunction, IReporter)
    done.append(1)


def _is_skipped(obj) -> bool:
    """Return True if the given object has been marked with @unittest.skip."""
    return bool(getattr(obj, "__unittest_skip__", False))



import asyncio

import pytest


class CustomSelectorLoop(asyncio.SelectorEventLoop):
    """A subclass with no overrides, just to test for presence."""


loop = CustomSelectorLoop()


@pytest.fixture(scope="module")
def event_loop():
    """Create an instance of the default event loop for each test case."""
    yield loop
    loop.close()



# -*- coding: utf-8 -*-
"""
test_related_events.py
~~~~~~~~~~~~~~~~~~~~~~

Specific tests to validate the "related events" logic used by certain events
inside hyper-h2.
"""
import h2.config
import h2.connection
import h2.events


class TestRelatedEvents(object):
    """
    Related events correlate all those events that happen on a single frame.
    """
    example_request_headers = [
        (':authority', 'example.com'),
        (':path', '/'),
        (':scheme', 'https'),
        (':method', 'GET'),
    ]

    example_response_headers = [
        (':status', '200'),
        ('server', 'fake-serv/0.1.0')
    ]

    informational_response_headers = [
        (':status', '100'),
        ('server', 'fake-serv/0.1.0')
    ]

    example_trailers = [
        ('another', 'field'),
    ]

    server_config = h2.config.H2Configuration(client_side=False)

    def test_request_received_related_all(self, frame_factory):
        """
        RequestReceived has two possible related events: PriorityUpdated and
        StreamEnded, all fired when a single HEADERS frame is received.
        """
        c = h2.connection.H2Connection(config=self.server_config)
        c.initiate_connection()
        c.receive_data(frame_factory.preamble())

        input_frame = frame_factory.build_headers_frame(
            headers=self.example_request_headers,
            flags=['END_STREAM', 'PRIORITY'],
            stream_weight=15,
            depends_on=0,
            exclusive=False,
        )
        events = c.receive_data(input_frame.serialize())

        assert len(events) == 3
        base_event = events[0]
        other_events = events[1:]

        assert base_event.stream_ended in other_events
        assert isinstance(base_event.stream_ended, h2.events.StreamEnded)
        assert base_event.priority_updated in other_events
        assert isinstance(
            base_event.priority_updated, h2.events.PriorityUpdated
        )

    def test_request_received_related_priority(self, frame_factory):
        """
        RequestReceived can be related to PriorityUpdated.
        """
        c = h2.connection.H2Connection(config=self.server_config)
        c.initiate_connection()
        c.receive_data(frame_factory.preamble())

        input_frame = frame_factory.build_headers_frame(
            headers=self.example_request_headers,
            flags=['PRIORITY'],
            stream_weight=15,
            depends_on=0,
            exclusive=False,
        )
        events = c.receive_data(input_frame.serialize())

        assert len(events) == 2
        base_event = events[0]
        priority_updated_event = events[1]

        assert base_event.priority_updated is priority_updated_event
        assert base_event.stream_ended is None
        assert isinstance(
            base_event.priority_updated, h2.events.PriorityUpdated
        )

    def test_request_received_related_stream_ended(self, frame_factory):
        """
        RequestReceived can be related to StreamEnded.
        """
        c = h2.connection.H2Connection(config=self.server_config)
        c.initiate_connection()
        c.receive_data(frame_factory.preamble())

        input_frame = frame_factory.build_headers_frame(
            headers=self.example_request_headers,
            flags=['END_STREAM'],
        )
        events = c.receive_data(input_frame.serialize())

        assert len(events) == 2
        base_event = events[0]
        stream_ended_event = events[1]

        assert base_event.stream_ended is stream_ended_event
        assert base_event.priority_updated is None
        assert isinstance(base_event.stream_ended, h2.events.StreamEnded)

    def test_response_received_related_nothing(self, frame_factory):
        """
        ResponseReceived is ordinarily related to no events.
        """
        c = h2.connection.H2Connection()
        c.initiate_connection()
        c.send_headers(stream_id=1, headers=self.example_request_headers)

        input_frame = frame_factory.build_headers_frame(
            headers=self.example_response_headers,
        )
        events = c.receive_data(input_frame.serialize())

        assert len(events) == 1
        base_event = events[0]

        assert base_event.stream_ended is None
        assert base_event.priority_updated is None

    def test_response_received_related_all(self, frame_factory):
        """
        ResponseReceived has two possible related events: PriorityUpdated and
        StreamEnded, all fired when a single HEADERS frame is received.
        """
        c = h2.connection.H2Connection()
        c.initiate_connection()
        c.send_headers(stream_id=1, headers=self.example_request_headers)

        input_frame = frame_factory.build_headers_frame(
            headers=self.example_response_headers,
            flags=['END_STREAM', 'PRIORITY'],
            stream_weight=15,
            depends_on=0,
            exclusive=False,
        )
        events = c.receive_data(input_frame.serialize())

        assert len(events) == 3
        base_event = events[0]
        other_events = events[1:]

        assert base_event.stream_ended in other_events
        assert isinstance(base_event.stream_ended, h2.events.StreamEnded)
        assert base_event.priority_updated in other_events
        assert isinstance(
            base_event.priority_updated, h2.events.PriorityUpdated
        )

    def test_response_received_related_priority(self, frame_factory):
        """
        ResponseReceived can be related to PriorityUpdated.
        """
        c = h2.connection.H2Connection()
        c.initiate_connection()
        c.send_headers(stream_id=1, headers=self.example_request_headers)

        input_frame = frame_factory.build_headers_frame(
            headers=self.example_response_headers,
            flags=['PRIORITY'],
            stream_weight=15,
            depends_on=0,
            exclusive=False,
        )
        events = c.receive_data(input_frame.serialize())

        assert len(events) == 2
        base_event = events[0]
        priority_updated_event = events[1]

        assert base_event.priority_updated is priority_updated_event
        assert base_event.stream_ended is None
        assert isinstance(
            base_event.priority_updated, h2.events.PriorityUpdated
        )

    def test_response_received_related_stream_ended(self, frame_factory):
        """
        ResponseReceived can be related to StreamEnded.
        """
        c = h2.connection.H2Connection()
        c.initiate_connection()
        c.send_headers(stream_id=1, headers=self.example_request_headers)

        input_frame = frame_factory.build_headers_frame(
            headers=self.example_response_headers,
            flags=['END_STREAM'],
        )
        events = c.receive_data(input_frame.serialize())

        assert len(events) == 2
        base_event = events[0]
        stream_ended_event = events[1]

        assert base_event.stream_ended is stream_ended_event
        assert base_event.priority_updated is None
        assert isinstance(base_event.stream_ended, h2.events.StreamEnded)

    def test_trailers_received_related_all(self, frame_factory):
        """
        TrailersReceived has two possible related events: PriorityUpdated and
        StreamEnded, all fired when a single HEADERS frame is received.
        """
        c = h2.connection.H2Connection()
        c.initiate_connection()
        c.send_headers(stream_id=1, headers=self.example_request_headers)

        f = frame_factory.build_headers_frame(
            headers=self.example_response_headers,
        )
        c.receive_data(f.serialize())

        input_frame = frame_factory.build_headers_frame(
            headers=self.example_trailers,
            flags=['END_STREAM', 'PRIORITY'],
            stream_weight=15,
            depends_on=0,
            exclusive=False,
        )
        events = c.receive_data(input_frame.serialize())

        assert len(events) == 3
        base_event = events[0]
        other_events = events[1:]

        assert base_event.stream_ended in other_events
        assert isinstance(base_event.stream_ended, h2.events.StreamEnded)
        assert base_event.priority_updated in other_events
        assert isinstance(
            base_event.priority_updated, h2.events.PriorityUpdated
        )

    def test_trailers_received_related_stream_ended(self, frame_factory):
        """
        TrailersReceived can be related to StreamEnded by itself.
        """
        c = h2.connection.H2Connection()
        c.initiate_connection()
        c.send_headers(stream_id=1, headers=self.example_request_headers)

        f = frame_factory.build_headers_frame(
            headers=self.example_response_headers,
        )
        c.receive_data(f.serialize())

        input_frame = frame_factory.build_headers_frame(
            headers=self.example_trailers,
            flags=['END_STREAM'],
        )
        events = c.receive_data(input_frame.serialize())

        assert len(events) == 2
        base_event = events[0]
        stream_ended_event = events[1]

        assert base_event.stream_ended is stream_ended_event
        assert base_event.priority_updated is None
        assert isinstance(base_event.stream_ended, h2.events.StreamEnded)

    def test_informational_response_related_nothing(self, frame_factory):
        """
        InformationalResponseReceived in the standard case is related to
        nothing.
        """
        c = h2.connection.H2Connection()
        c.initiate_connection()
        c.send_headers(stream_id=1, headers=self.example_request_headers)

        input_frame = frame_factory.build_headers_frame(
            headers=self.informational_response_headers,
        )
        events = c.receive_data(input_frame.serialize())

        assert len(events) == 1
        base_event = events[0]

        assert base_event.priority_updated is None

    def test_informational_response_received_related_all(self, frame_factory):
        """
        InformationalResponseReceived has one possible related event:
        PriorityUpdated, fired when a single HEADERS frame is received.
        """
        c = h2.connection.H2Connection()
        c.initiate_connection()
        c.send_headers(stream_id=1, headers=self.example_request_headers)

        input_frame = frame_factory.build_headers_frame(
            headers=self.informational_response_headers,
            flags=['PRIORITY'],
            stream_weight=15,
            depends_on=0,
            exclusive=False,
        )
        events = c.receive_data(input_frame.serialize())

        assert len(events) == 2
        base_event = events[0]
        priority_updated_event = events[1]

        assert base_event.priority_updated is priority_updated_event
        assert isinstance(
            base_event.priority_updated, h2.events.PriorityUpdated
        )

    def test_data_received_normally_relates_to_nothing(self, frame_factory):
        """
        A plain DATA frame leads to DataReceieved with no related events.
        """
        c = h2.connection.H2Connection()
        c.initiate_connection()
        c.send_headers(stream_id=1, headers=self.example_request_headers)

        f = frame_factory.build_headers_frame(
            headers=self.example_response_headers,
        )
        c.receive_data(f.serialize())

        input_frame = frame_factory.build_data_frame(
            data=b'some data',
        )
        events = c.receive_data(input_frame.serialize())

        assert len(events) == 1
        base_event = events[0]

        assert base_event.stream_ended is None

    def test_data_received_related_stream_ended(self, frame_factory):
        """
        DataReceived can be related to StreamEnded by itself.
        """
        c = h2.connection.H2Connection()
        c.initiate_connection()
        c.send_headers(stream_id=1, headers=self.example_request_headers)

        f = frame_factory.build_headers_frame(
            headers=self.example_response_headers,
        )
        c.receive_data(f.serialize())

        input_frame = frame_factory.build_data_frame(
            data=b'some data',
            flags=['END_STREAM'],
        )
        events = c.receive_data(input_frame.serialize())

        assert len(events) == 2
        base_event = events[0]
        stream_ended_event = events[1]

        assert base_event.stream_ended is stream_ended_event
        assert isinstance(base_event.stream_ended, h2.events.StreamEnded)



from __future__ import absolute_import, division, unicode_literals

from collections import OrderedDict

import pytest

import html5lib
from html5lib.filters.alphabeticalattributes import Filter
from html5lib.serializer import HTMLSerializer


@pytest.mark.parametrize('msg, attrs, expected_attrs', [
    (
        'no attrs',
        {},
        {}
    ),
    (
        'one attr',
        {(None, 'alt'): 'image'},
        OrderedDict([((None, 'alt'), 'image')])
    ),
    (
        'multiple attrs',
        {
            (None, 'src'): 'foo',
            (None, 'alt'): 'image',
            (None, 'style'): 'border: 1px solid black;'
        },
        OrderedDict([
            ((None, 'alt'), 'image'),
            ((None, 'src'), 'foo'),
            ((None, 'style'), 'border: 1px solid black;')
        ])
    ),
])
def test_alphabetizing(msg, attrs, expected_attrs):
    tokens = [{'type': 'StartTag', 'name': 'img', 'data': attrs}]
    output_tokens = list(Filter(tokens))

    attrs = output_tokens[0]['data']
    assert attrs == expected_attrs


def test_with_different_namespaces():
    tokens = [{
        'type': 'StartTag',
        'name': 'pattern',
        'data': {
            (None, 'id'): 'patt1',
            ('http://www.w3.org/1999/xlink', 'href'): '#patt2'
        }
    }]
    output_tokens = list(Filter(tokens))

    attrs = output_tokens[0]['data']
    assert attrs == OrderedDict([
        ((None, 'id'), 'patt1'),
        (('http://www.w3.org/1999/xlink', 'href'), '#patt2')
    ])


def test_with_serializer():
    """Verify filter works in the context of everything else"""
    parser = html5lib.HTMLParser()
    dom = parser.parseFragment('<svg><pattern xlink:href="#patt2" id="patt1"></svg>')
    walker = html5lib.getTreeWalker('etree')
    ser = HTMLSerializer(
        alphabetical_attributes=True,
        quote_attr_values='always'
    )

    # FIXME(willkg): The "xlink" namespace gets dropped by the serializer. When
    # that gets fixed, we can fix this expected result.
    assert (
        ser.render(walker(dom)) ==
        '<svg><pattern id="patt1" href="#patt2"></pattern></svg>'
    )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import os
import sys

import six


class NullTerminal(object):
    """Replacement for `blessed.Terminal()` that does no formatting."""

    number_of_colors = 0
    width = 0
    height = 0

    def __init__(self, stream=None, **kwargs):
        self.stream = stream or sys.__stdout__
        try:
            self.is_a_tty = os.isatty(self.stream.fileno())
        except Exception:
            self.is_a_tty = False

    class NullCallableString(six.text_type):
        """A dummy callable Unicode stolen from blessings"""

        def __new__(cls):
            new = six.text_type.__new__(cls, "")
            return new

        def __call__(self, *args):
            if len(args) != 1 or isinstance(args[0], int):
                return ""
            return args[0]

    def __getattr__(self, attr):
        return self.NullCallableString()


def Terminal(raises=False, disable_styling=False, **kwargs):
    if disable_styling:
        return NullTerminal(**kwargs)
    try:
        import blessed
    except Exception:
        if raises:
            raise
        return NullTerminal(**kwargs)
    return blessed.Terminal(**kwargs)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from mach.decorators import Command


def is_true(cls):
    return True


def is_false(cls):
    return False


@Command("cmd_condition_true", category="testing", conditions=[is_true])
def run_condition_true(self, command_context):
    pass


@Command("cmd_condition_false", category="testing", conditions=[is_false])
def run_condition_false(self, command_context):
    pass


@Command(
    "cmd_condition_true_and_false", category="testing", conditions=[is_true, is_false]
)
def run_condition_true_and_false(self, command_context):
    pass


def is_ctx_foo(cls):
    """Foo must be true"""
    return cls._mach_context.foo


def is_ctx_bar(cls):
    """Bar must be true"""
    return cls._mach_context.bar


@Command("cmd_foo_ctx", category="testing", conditions=[is_ctx_foo])
def run_foo_ctx(self, command_context):
    pass


@Command("cmd_bar_ctx", category="testing", conditions=[is_ctx_bar])
def run_bar_ctx(self, command_context):
    pass


@Command("cmd_foobar_ctx", category="testing", conditions=[is_ctx_foo, is_ctx_bar])
def run_foobar_ctx(self, command_context):
    pass



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import os
import sys
from unittest import skipIf

from marionette_driver.addons import Addons, AddonInstallException
from marionette_driver.errors import UnknownException
from marionette_harness import MarionetteTestCase


here = os.path.abspath(os.path.dirname(__file__))


class TestAddons(MarionetteTestCase):
    def setUp(self):
        super(TestAddons, self).setUp()

        self.addons = Addons(self.marionette)
        self.preinstalled_addons = self.all_addon_ids

    def tearDown(self):
        self.reset_addons()

        super(TestAddons, self).tearDown()

    @property
    def all_addon_ids(self):
        with self.marionette.using_context("chrome"):
            addons = self.marionette.execute_async_script(
                """
              const [resolve] = arguments;
              const { AddonManager } = ChromeUtils.importESModule(
                "resource://gre/modules/AddonManager.sys.mjs"
              );

              async function getAllAddons() {
                const addons = await AddonManager.getAllAddons();
                const ids = addons.map(x => x.id);
                resolve(ids);
              }

              getAllAddons();
            """
            )

        return set(addons)

    def reset_addons(self):
        with self.marionette.using_context("chrome"):
            for addon in self.all_addon_ids - self.preinstalled_addons:
                addon_id = self.marionette.execute_async_script(
                    """
                  const [addonId, resolve] = arguments;
                  const { AddonManager } = ChromeUtils.importESModule(
                    "resource://gre/modules/AddonManager.sys.mjs"
                  );

                  async function uninstall() {
                    const addon = await AddonManager.getAddonByID(addonId);
                    addon.uninstall();
                    resolve(addon.id);
                  }

                  uninstall();
                """,
                    script_args=(addon,),
                )
                self.assertEqual(
                    addon_id, addon, msg="Failed to uninstall {}".format(addon)
                )

    def test_temporary_install_and_remove_unsigned_addon(self):
        addon_path = os.path.join(here, "webextension-unsigned.xpi")

        addon_id = self.addons.install(addon_path, temp=True)
        self.assertIn(addon_id, self.all_addon_ids)
        self.assertEqual(addon_id, "{d3e7c1f1-2e35-4a49-89fe-9f46eb8abf0a}")

        self.addons.uninstall(addon_id)
        self.assertNotIn(addon_id, self.all_addon_ids)

    def test_temporary_install_invalid_addon(self):
        addon_path = os.path.join(here, "webextension-invalid.xpi")

        with self.assertRaises(AddonInstallException):
            self.addons.install(addon_path, temp=True)
        self.assertNotIn("{d3e7c1f1-2e35-4a49-89fe-9f46eb8abf0a}", self.all_addon_ids)

    def test_install_and_remove_signed_addon(self):
        addon_path = os.path.join(here, "amosigned.xpi")

        addon_id = self.addons.install(addon_path)
        self.assertIn(addon_id, self.all_addon_ids)
        self.assertEqual(addon_id, "amosigned-xpi@tests.mozilla.org")

        self.addons.uninstall(addon_id)
        self.assertNotIn(addon_id, self.all_addon_ids)

    def test_install_invalid_addon(self):
        addon_path = os.path.join(here, "webextension-invalid.xpi")

        with self.assertRaises(AddonInstallException):
            self.addons.install(addon_path)
        self.assertNotIn("{d3e7c1f1-2e35-4a49-89fe-9f46eb8abf0a}", self.all_addon_ids)

    def test_install_unsigned_addon_fails(self):
        addon_path = os.path.join(here, "webextension-unsigned.xpi")

        with self.assertRaises(AddonInstallException):
            self.addons.install(addon_path)

    def test_install_nonexistent_addon(self):
        addon_path = os.path.join(here, "does-not-exist.xpi")

        with self.assertRaises(AddonInstallException):
            self.addons.install(addon_path)

    def test_install_with_relative_path(self):
        with self.assertRaises(AddonInstallException):
            self.addons.install("webextension.xpi")

    @skipIf(sys.platform != "win32", "Only makes sense on Windows")
    def test_install_mixed_separator_windows(self):
        # Ensure the base path has only \
        addon_path = here.replace("/", "\\")
        addon_path += "/amosigned.xpi"

        addon_id = self.addons.install(addon_path, temp=True)
        self.assertIn(addon_id, self.all_addon_ids)
        self.assertEqual(addon_id, "amosigned-xpi@tests.mozilla.org")

        self.addons.uninstall(addon_id)
        self.assertNotIn(addon_id, self.all_addon_ids)

    def test_uninstall_nonexistent_addon(self):
        with self.assertRaises(UnknownException):
            self.addons.uninstall("i-do-not-exist-as-an-id")



#!/usr/bin/env python
# coding=UTF-8

import mozunit
import pytest
from conftest import fspath


def test_no_dump_files(check_for_crashes):
    """Test that check_for_crashes returns 0 if no dumps are present."""
    assert 0 == check_for_crashes()


@pytest.mark.parametrize("minidump_files", [3], indirect=True)
def test_dump_count(check_for_crashes, minidump_files):
    """Test that check_for_crashes returns the number of crash dumps."""
    assert 3 == check_for_crashes()


def test_dump_directory_unicode(request, check_for_crashes, tmpdir, capsys):
    """Test that check_for_crashes can handle unicode in dump_directory."""
    from conftest import minidump_files

    tmpdir = tmpdir.ensure("🍪", dir=1)
    minidump_files = minidump_files(request, tmpdir)

    assert 1 == check_for_crashes(dump_directory=fspath(tmpdir), quiet=False)

    out, _ = capsys.readouterr()
    assert fspath(minidump_files[0]["dmp"]) in out
    assert "🍪" in out


def test_test_name_unicode(check_for_crashes, minidump_files, capsys):
    """Test that check_for_crashes can handle unicode in dump_directory."""
    assert 1 == check_for_crashes(test_name="🍪", quiet=False)

    out, err = capsys.readouterr()
    assert "| 🍪" in out


if __name__ == "__main__":
    mozunit.main()



from marionette_driver.by import By
from marionette_driver.expected import element_present
from marionette_driver import errors
from marionette_driver.marionette import Alert
from marionette_driver.wait import Wait

from marionette_harness import MarionetteTestCase, parameterized, WindowManagerMixin


class TestModalDialogs(WindowManagerMixin, MarionetteTestCase):
    def setUp(self):
        super(TestModalDialogs, self).setUp()
        self.new_tab = self.open_tab()
        self.marionette.switch_to_window(self.new_tab)

        self.http_auth_pref = (
            "network.auth.non-web-content-triggered-resources-http-auth-allow"
        )

    def tearDown(self):
        # Ensure to close all possible remaining tab modal dialogs
        try:
            while True:
                alert = self.marionette.switch_to_alert()
                alert.dismiss()
        except errors.NoAlertPresentException:
            pass

        self.close_all_tabs()
        self.close_all_windows()

        super(TestModalDialogs, self).tearDown()

    @property
    def alert_present(self):
        try:
            Alert(self.marionette).text
            return True
        except errors.NoAlertPresentException:
            return False

    def wait_for_alert(self, timeout=None):
        Wait(self.marionette, timeout=timeout).until(lambda _: self.alert_present)

    def open_custom_prompt(self, modal_type, delay=0):
        browsing_context_id = self.marionette.execute_script(
            """
            return window.browsingContext.id;
        """,
            sandbox="system",
        )

        with self.marionette.using_context("chrome"):
            self.marionette.execute_script(
                """
                const [ modalType, browsingContextId, delay ] = arguments;

                const modalTypes = {
                  1: Services.prompt.MODAL_TYPE_CONTENT,
                  2: Services.prompt.MODAL_TYPE_TAB,
                  3: Services.prompt.MODAL_TYPE_WINDOW,
                  4: Services.prompt.MODAL_TYPE_INTERNAL_WINDOW,
                }

                window.setTimeout(() => {
                  Services.prompt.alertBC(
                    BrowsingContext.get(browsingContextId),
                    modalTypes[modalType],
                    "title",
                    "text"
                  );
                }, delay);
            """,
                script_args=(modal_type, browsing_context_id, delay * 1000),
            )

    @parameterized("content", 1)
    @parameterized("tab", 2)
    @parameterized("window", 3)
    @parameterized("internal_window", 4)
    def test_detect_modal_type_in_current_tab_for_type(self, type):
        self.open_custom_prompt(type)
        self.wait_for_alert()

        self.assertTrue(self.alert_present)

        # Restart the session to ensure we still find the formerly left-open dialog.
        self.marionette.delete_session()
        self.marionette.start_session()

        alert = self.marionette.switch_to_alert()
        alert.dismiss()

    @parameterized("content", 1)
    @parameterized("tab", 2)
    def test_dont_detect_content_and_tab_modal_type_in_another_tab_for_type(self, type):
        self.open_custom_prompt(type, delay=0.25)

        self.marionette.switch_to_window(self.start_tab)
        with self.assertRaises(errors.TimeoutException):
            self.wait_for_alert(2)

        self.marionette.switch_to_window(self.new_tab)
        alert = self.marionette.switch_to_alert()
        alert.dismiss()

    @parameterized("window", 3)
    @parameterized("internal_window", 4)
    def test_detect_window_modal_type_in_another_tab_for_type(self, type):
        self.open_custom_prompt(type, delay=0.25)

        self.marionette.switch_to_window(self.start_tab)
        self.wait_for_alert()

        alert = self.marionette.switch_to_alert()
        alert.dismiss()

        self.marionette.switch_to_window(self.new_tab)
        self.assertFalse(self.alert_present)

    @parameterized("window", 3)
    @parameterized("internal_window", 4)
    def test_detect_window_modal_type_in_another_window_for_type(self, type):
        self.new_window = self.open_window()

        self.marionette.switch_to_window(self.new_window)

        self.open_custom_prompt(type, delay=0.25)

        self.marionette.switch_to_window(self.new_tab)
        with self.assertRaises(errors.TimeoutException):
            self.wait_for_alert(2)

        self.marionette.switch_to_window(self.new_window)
        alert = self.marionette.switch_to_alert()
        alert.dismiss()

        self.marionette.switch_to_window(self.new_tab)
        self.assertFalse(self.alert_present)

    def test_http_auth_dismiss(self):
        with self.marionette.using_prefs({self.http_auth_pref: True}):
            self.marionette.navigate(self.marionette.absolute_url("http_auth"))
            self.wait_for_alert(timeout=self.marionette.timeout.page_load)

            alert = self.marionette.switch_to_alert()
            alert.dismiss()

            status = Wait(
                self.marionette, timeout=self.marionette.timeout.page_load
            ).until(element_present(By.ID, "status"))
            self.assertEqual(status.text, "restricted")

    def test_http_auth_send_keys(self):
        with self.marionette.using_prefs({self.http_auth_pref: True}):
            self.marionette.navigate(self.marionette.absolute_url("http_auth"))
            self.wait_for_alert(timeout=self.marionette.timeout.page_load)

            alert = self.marionette.switch_to_alert()
            with self.assertRaises(errors.UnsupportedOperationException):
                alert.send_keys("foo")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import os
import threading
from time import sleep

import mozrunner
import pytest
from moztest.selftest import fixtures


@pytest.fixture(scope="session")
def get_binary():
    if "BROWSER_PATH" in os.environ:
        os.environ["GECKO_BINARY_PATH"] = os.environ["BROWSER_PATH"]

    def inner(app):
        if app not in ("chrome", "chromium", "firefox"):
            pytest.xfail(reason="{} support not implemented".format(app))

        if app == "firefox":
            binary = fixtures.binary()
        elif app == "chrome":
            binary = os.environ.get("CHROME_BINARY_PATH")
        elif app == "chromium":
            binary = os.environ.get("CHROMIUM_BINARY_PATH")

        if not binary:
            pytest.skip("could not find a {} binary".format(app))
        return binary

    return inner


@pytest.fixture(params=["firefox", "chrome", "chromium"])
def runner(request, get_binary):
    app = request.param
    binary = get_binary(app)

    cmdargs = ["--headless"]
    if app in ["chrome", "chromium"]:
        # prevents headless chromium from exiting after loading the page
        cmdargs.append("--remote-debugging-port=9222")
        # only needed on Windows, but no harm in specifying it everywhere
        cmdargs.append("--disable-gpu")
    runner = mozrunner.runners[app](binary, cmdargs=cmdargs)
    runner.app = app
    yield runner
    runner.stop()


class RunnerThread(threading.Thread):
    def __init__(self, runner, start=False, timeout=1):
        threading.Thread.__init__(self)
        self.runner = runner
        self.timeout = timeout
        self.do_start = start

    def run(self):
        sleep(self.timeout)
        if self.do_start:
            self.runner.start()
        else:
            self.runner.stop()


@pytest.fixture
def create_thread():
    threads = []

    def inner(*args, **kwargs):
        thread = RunnerThread(*args, **kwargs)
        threads.append(thread)
        return thread

    yield inner

    for thread in threads:
        thread.join()



import argparse
import os
import re
import sys
from pathlib import Path
from typing import Optional

import pytest
from _pytest.config import ExitCode
from _pytest.config import UsageError
from _pytest.main import resolve_collection_argument
from _pytest.main import validate_basetemp
from _pytest.pytester import Pytester


@pytest.mark.parametrize(
    "ret_exc",
    (
        pytest.param((None, ValueError)),
        pytest.param((42, SystemExit)),
        pytest.param((False, SystemExit)),
    ),
)
def test_wrap_session_notify_exception(ret_exc, pytester: Pytester) -> None:
    returncode, exc = ret_exc
    c1 = pytester.makeconftest(
        """
        import pytest

        def pytest_sessionstart():
            raise {exc}("boom")

        def pytest_internalerror(excrepr, excinfo):
            returncode = {returncode!r}
            if returncode is not False:
                pytest.exit("exiting after %s..." % excinfo.typename, returncode={returncode!r})
    """.format(
            returncode=returncode, exc=exc.__name__
        )
    )
    result = pytester.runpytest()
    if returncode:
        assert result.ret == returncode
    else:
        assert result.ret == ExitCode.INTERNAL_ERROR
    assert result.stdout.lines[0] == "INTERNALERROR> Traceback (most recent call last):"

    end_lines = (
        result.stdout.lines[-4:]
        if sys.version_info >= (3, 11)
        else result.stdout.lines[-3:]
    )

    if exc == SystemExit:
        assert end_lines == [
            f'INTERNALERROR>   File "{c1}", line 4, in pytest_sessionstart',
            'INTERNALERROR>     raise SystemExit("boom")',
            *(
                ("INTERNALERROR>     ^^^^^^^^^^^^^^^^^^^^^^^^",)
                if sys.version_info >= (3, 11)
                else ()
            ),
            "INTERNALERROR> SystemExit: boom",
        ]
    else:
        assert end_lines == [
            f'INTERNALERROR>   File "{c1}", line 4, in pytest_sessionstart',
            'INTERNALERROR>     raise ValueError("boom")',
            *(
                ("INTERNALERROR>     ^^^^^^^^^^^^^^^^^^^^^^^^",)
                if sys.version_info >= (3, 11)
                else ()
            ),
            "INTERNALERROR> ValueError: boom",
        ]
    if returncode is False:
        assert result.stderr.lines == ["mainloop: caught unexpected SystemExit!"]
    else:
        assert result.stderr.lines == [f"Exit: exiting after {exc.__name__}..."]


@pytest.mark.parametrize("returncode", (None, 42))
def test_wrap_session_exit_sessionfinish(
    returncode: Optional[int], pytester: Pytester
) -> None:
    pytester.makeconftest(
        """
        import pytest
        def pytest_sessionfinish():
            pytest.exit(reason="exit_pytest_sessionfinish", returncode={returncode})
    """.format(
            returncode=returncode
        )
    )
    result = pytester.runpytest()
    if returncode:
        assert result.ret == returncode
    else:
        assert result.ret == ExitCode.NO_TESTS_COLLECTED
    assert result.stdout.lines[-1] == "collected 0 items"
    assert result.stderr.lines == ["Exit: exit_pytest_sessionfinish"]


@pytest.mark.parametrize("basetemp", ["foo", "foo/bar"])
def test_validate_basetemp_ok(tmp_path, basetemp, monkeypatch):
    monkeypatch.chdir(str(tmp_path))
    validate_basetemp(tmp_path / basetemp)


@pytest.mark.parametrize("basetemp", ["", ".", ".."])
def test_validate_basetemp_fails(tmp_path, basetemp, monkeypatch):
    monkeypatch.chdir(str(tmp_path))
    msg = "basetemp must not be empty, the current working directory or any parent directory of it"
    with pytest.raises(argparse.ArgumentTypeError, match=msg):
        if basetemp:
            basetemp = tmp_path / basetemp
        validate_basetemp(basetemp)


def test_validate_basetemp_integration(pytester: Pytester) -> None:
    result = pytester.runpytest("--basetemp=.")
    result.stderr.fnmatch_lines("*basetemp must not be*")


class TestResolveCollectionArgument:
    @pytest.fixture
    def invocation_path(self, pytester: Pytester) -> Path:
        pytester.syspathinsert(pytester.path / "src")
        pytester.chdir()

        pkg = pytester.path.joinpath("src/pkg")
        pkg.mkdir(parents=True)
        pkg.joinpath("__init__.py").touch()
        pkg.joinpath("test.py").touch()
        return pytester.path

    def test_file(self, invocation_path: Path) -> None:
        """File and parts."""
        assert resolve_collection_argument(invocation_path, "src/pkg/test.py") == (
            invocation_path / "src/pkg/test.py",
            [],
        )
        assert resolve_collection_argument(invocation_path, "src/pkg/test.py::") == (
            invocation_path / "src/pkg/test.py",
            [""],
        )
        assert resolve_collection_argument(
            invocation_path, "src/pkg/test.py::foo::bar"
        ) == (invocation_path / "src/pkg/test.py", ["foo", "bar"])
        assert resolve_collection_argument(
            invocation_path, "src/pkg/test.py::foo::bar::"
        ) == (invocation_path / "src/pkg/test.py", ["foo", "bar", ""])

    def test_dir(self, invocation_path: Path) -> None:
        """Directory and parts."""
        assert resolve_collection_argument(invocation_path, "src/pkg") == (
            invocation_path / "src/pkg",
            [],
        )

        with pytest.raises(
            UsageError, match=r"directory argument cannot contain :: selection parts"
        ):
            resolve_collection_argument(invocation_path, "src/pkg::")

        with pytest.raises(
            UsageError, match=r"directory argument cannot contain :: selection parts"
        ):
            resolve_collection_argument(invocation_path, "src/pkg::foo::bar")

    def test_pypath(self, invocation_path: Path) -> None:
        """Dotted name and parts."""
        assert resolve_collection_argument(
            invocation_path, "pkg.test", as_pypath=True
        ) == (invocation_path / "src/pkg/test.py", [])
        assert resolve_collection_argument(
            invocation_path, "pkg.test::foo::bar", as_pypath=True
        ) == (invocation_path / "src/pkg/test.py", ["foo", "bar"])
        assert resolve_collection_argument(invocation_path, "pkg", as_pypath=True) == (
            invocation_path / "src/pkg",
            [],
        )

        with pytest.raises(
            UsageError, match=r"package argument cannot contain :: selection parts"
        ):
            resolve_collection_argument(
                invocation_path, "pkg::foo::bar", as_pypath=True
            )

    def test_parametrized_name_with_colons(self, invocation_path: Path) -> None:
        ret = resolve_collection_argument(
            invocation_path, "src/pkg/test.py::test[a::b]"
        )
        assert ret == (invocation_path / "src/pkg/test.py", ["test[a::b]"])

    def test_does_not_exist(self, invocation_path: Path) -> None:
        """Given a file/module that does not exist raises UsageError."""
        with pytest.raises(
            UsageError, match=re.escape("file or directory not found: foobar")
        ):
            resolve_collection_argument(invocation_path, "foobar")

        with pytest.raises(
            UsageError,
            match=re.escape(
                "module or package not found: foobar (missing __init__.py?)"
            ),
        ):
            resolve_collection_argument(invocation_path, "foobar", as_pypath=True)

    def test_absolute_paths_are_resolved_correctly(self, invocation_path: Path) -> None:
        """Absolute paths resolve back to absolute paths."""
        full_path = str(invocation_path / "src")
        assert resolve_collection_argument(invocation_path, full_path) == (
            Path(os.path.abspath("src")),
            [],
        )

        # ensure full paths given in the command-line without the drive letter resolve
        # to the full path correctly (#7628)
        drive, full_path_without_drive = os.path.splitdrive(full_path)
        assert resolve_collection_argument(
            invocation_path, full_path_without_drive
        ) == (Path(os.path.abspath("src")), [])


def test_module_full_path_without_drive(pytester: Pytester) -> None:
    """Collect and run test using full path except for the drive letter (#7628).

    Passing a full path without a drive letter would trigger a bug in legacy_path
    where it would keep the full path without the drive letter around, instead of resolving
    to the full path, resulting in fixtures node ids not matching against test node ids correctly.
    """
    pytester.makepyfile(
        **{
            "project/conftest.py": """
                import pytest
                @pytest.fixture
                def fix(): return 1
            """,
        }
    )

    pytester.makepyfile(
        **{
            "project/tests/dummy_test.py": """
                def test(fix):
                    assert fix == 1
            """
        }
    )
    fn = pytester.path.joinpath("project/tests/dummy_test.py")
    assert fn.is_file()

    drive, path = os.path.splitdrive(str(fn))

    result = pytester.runpytest(path, "-v")
    result.stdout.fnmatch_lines(
        [
            os.path.join("project", "tests", "dummy_test.py") + "::test PASSED *",
            "* 1 passed in *",
        ]
    )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import sys
from io import StringIO

import mozunit
import pytest

from mozterm import Terminal
from mozterm.widgets import Footer


@pytest.fixture
def terminal():
    blessed = pytest.importorskip("blessed")

    kind = "xterm-256color"
    try:
        term = Terminal(stream=StringIO(), force_styling=True, kind=kind)
    except blessed.curses.error:
        pytest.skip("terminal '{}' not found".format(kind))

    return term


@pytest.mark.skipif(
    not sys.platform.startswith("win"),
    reason="Only do ANSI Escape Sequence comparisons on Windows.",
)
def test_footer(terminal):
    footer = Footer(terminal=terminal)
    footer.write(
        [
            ("bright_black", "foo"),
            ("green", "bar"),
        ]
    )
    value = terminal.stream.getvalue()
    expected = "\x1b7\x1b[90mfoo\x1b(B\x1b[m \x1b[32mbar\x1b(B\x1b[m\x1b8"
    assert value == expected

    footer.clear()
    value = terminal.stream.getvalue()[len(value) :]
    expected = "\x1b[1G\x1b[K"
    assert value == expected


if __name__ == "__main__":
    mozunit.main()



#!/usr/bin/env python

import os
import signal
import sys

import mozinfo
import mozunit
import proctest
from mozprocess import processhandler

here = os.path.dirname(os.path.abspath(__file__))


class ProcTestWait(proctest.ProcTest):
    """Class to test process waits and timeouts"""

    def test_normal_finish(self):
        """Process is started, runs to completion while we wait for it"""

        p = processhandler.ProcessHandler(
            [self.python, self.proclaunch, "process_normal_finish.ini"], cwd=here
        )
        p.run()
        p.wait()

        self.determine_status(p)

    def test_wait(self):
        """Process is started runs to completion while we wait indefinitely"""

        p = processhandler.ProcessHandler(
            [self.python, self.proclaunch, "process_waittimeout_10s.ini"], cwd=here
        )
        p.run()
        p.wait()

        self.determine_status(p)

    def test_timeout(self):
        """Process is started, runs but we time out waiting on it
        to complete
        """
        myenv = None
        # On macosx1014, subprocess fails to find `six` when run with python3.
        # This ensures that subprocess first looks to sys.path to find `six`.
        # See https://bugzilla.mozilla.org/show_bug.cgi?id=1562083
        if sys.platform == "darwin" and sys.version_info[0] > 2:
            myenv = os.environ.copy()
            myenv["PYTHONPATH"] = ":".join(sys.path)

        p = processhandler.ProcessHandler(
            [self.python, self.proclaunch, "process_waittimeout.ini"],
            cwd=here,
            env=myenv,
        )
        p.run(timeout=10)
        p.wait()

        if mozinfo.isUnix:
            # process was killed, so returncode should be negative
            self.assertLess(p.proc.returncode, 0)

        self.determine_status(p, False, ["returncode", "didtimeout"])

    def test_waittimeout(self):
        """
        Process is started, then wait is called and times out.
        Process is still running and didn't timeout
        """
        p = processhandler.ProcessHandler(
            [self.python, self.proclaunch, "process_waittimeout_10s.ini"], cwd=here
        )

        p.run()
        p.wait(timeout=0)

        self.determine_status(p, True, ())

    def test_waitnotimeout(self):
        """Process is started, runs to completion before our wait times out"""
        p = processhandler.ProcessHandler(
            [self.python, self.proclaunch, "process_waittimeout_10s.ini"], cwd=here
        )
        p.run(timeout=30)
        p.wait()

        self.determine_status(p)

    def test_wait_twice_after_kill(self):
        """Bug 968718: Process is started and stopped. wait() twice afterward."""
        p = processhandler.ProcessHandler(
            [self.python, self.proclaunch, "process_waittimeout.ini"], cwd=here
        )
        p.run()
        p.kill()
        returncode1 = p.wait()
        returncode2 = p.wait()

        self.determine_status(p)

        # We killed the process, so the returncode should be non-zero
        if mozinfo.isWin:
            self.assertGreater(
                returncode2, 0, 'Positive returncode expected, got "%s"' % returncode2
            )
        else:
            self.assertLess(
                returncode2, 0, 'Negative returncode expected, got "%s"' % returncode2
            )
        self.assertEqual(
            returncode1, returncode2, "Expected both returncodes of wait() to be equal"
        )

    def test_wait_after_external_kill(self):
        """Process is killed externally, and poll() is called."""
        p = processhandler.ProcessHandler(
            [self.python, self.proclaunch, "process_normal_finish.ini"], cwd=here
        )
        p.run()
        os.kill(p.pid, signal.SIGTERM)
        returncode = p.wait()

        # We killed the process, so the returncode should be non-zero
        if mozinfo.isWin:
            self.assertEqual(
                returncode,
                signal.SIGTERM,
                'Positive returncode expected, got "%s"' % returncode,
            )
        else:
            self.assertEqual(
                returncode,
                -signal.SIGTERM,
                '%s expected, got "%s"' % (-signal.SIGTERM, returncode),
            )

        self.assertEqual(returncode, p.poll())

        self.determine_status(p)


if __name__ == "__main__":
    mozunit.main()



# SPDX-License-Identifier: MIT

"""
Tests for `__init_subclass__` related tests.

Python 3.6+ only.
"""

import pytest

import attr


@pytest.mark.parametrize("slots", [True, False])
def test_init_subclass_vanilla(slots):
    """
    `super().__init_subclass__` can be used if the subclass is not an attrs
    class both with dict and slotted classes.
    """

    @attr.s(slots=slots)
    class Base:
        def __init_subclass__(cls, param, **kw):
            super().__init_subclass__(**kw)
            cls.param = param

    class Vanilla(Base, param="foo"):
        pass

    assert "foo" == Vanilla().param


def test_init_subclass_attrs():
    """
    `__init_subclass__` works with attrs classes as long as slots=False.
    """

    @attr.s(slots=False)
    class Base:
        def __init_subclass__(cls, param, **kw):
            super().__init_subclass__(**kw)
            cls.param = param

    @attr.s
    class Attrs(Base, param="foo"):
        pass

    assert "foo" == Attrs().param
